from functools import partial

from PyQt5 import QtWidgets
from PyQt5 import QtGui
from PyQt5 import QtCore
from PyQt5.QtCore import Qt
from segment_folder import SegmentFolderWidget
from segment import segment_full_image

def add_network_menu(window, menu_bar):
    """ Not in use right now as training happens automatically when the 
        annotation is saved
    """
    network_menu = menu_bar.addMenu('Network')

    # start training
    start_training_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Start training', window)
    start_training_btn.triggered.connect(window.start_training)
    network_menu.addAction(start_training_btn)

    # stop training
    stop_training_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Stop training', window)
    stop_training_btn.triggered.connect(window.stop_training)
    network_menu.addAction(stop_training_btn)

    # segment folder
    segment_folder_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Segment folder', window)

    # Alt+S
    segment_action = QtWidgets.QAction(QtGui.QIcon(""), "Segment full image", window)
    segment_action.setShortcut("Alt+S")
    network_menu.addAction(segment_action)
    def seg_im(window):
        segment_full_image(window, overwrite=True)
    segment_action.triggered.connect(partial(seg_im, window))

    def show_segment_folder():
        window.segment_folder_widget = SegmentFolderWidget(window.sync_dir,
                                                         window.instruction_dir,
                                                         window.classes)
        window.segment_folder_widget.show()
    segment_folder_btn.triggered.connect(show_segment_folder)
    network_menu.addAction(segment_folder_btn)


def add_edit_menu(window, im_viewer, menu_bar, skip_fill=True):
    edit_menu = menu_bar.addMenu("Edit")

    # Undo
    undo_action = QtWidgets.QAction(QtGui.QIcon(""), "Undo", window)
    undo_action.setShortcut("Z")
    edit_menu.addAction(undo_action)
    undo_action.triggered.connect(im_viewer.scene.undo)

    # Redo
    redo_action = QtWidgets.QAction(QtGui.QIcon(""), "Redo", window)
    redo_action.setShortcut("Ctrl+Shift+Z")
    edit_menu.addAction(redo_action)
    redo_action.triggered.connect(im_viewer.scene.redo)

    # Save annotation
    save_annotation_action = QtWidgets.QAction(QtGui.QIcon(""), "Save annotation", window)
    save_annotation_action.setShortcut("Ctrl+Shift+Z")
    edit_menu.addAction(save_annotation_action)
    save_annotation_action.triggered.connect(im_viewer.parent.save_annotation)

    if not skip_fill:
        raise Exception('disabled')
        # skip the fill with the sagittal view, it's too annoying when this gets pressed by accident

        # Using alt key and clicking is slow when wanting to fill a 
        # closed foreground annotation with foreground and background with background.
        fill_slice_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                    'Fill slice', window)
        fill_slice_btn.setShortcut('G')
        fill_slice_btn.setStatusTip('Fill slice')
        fill_slice_btn.triggered.connect(im_viewer.fill_slice)
        edit_menu.addAction(fill_slice_btn)

    return edit_menu


def add_windows_menu(main_window):
    # contrast slider
    menu = main_window.menu_bar.addMenu("Windows")
    contrast_settings_action = QtWidgets.QAction(QtGui.QIcon(""), "Contrast settings", main_window)
    contrast_settings_action.setStatusTip('Show contrast settings')
    contrast_settings_action.triggered.connect(main_window.contrast_slider.show)
    menu.addAction(contrast_settings_action)

    show_sagittal_view_action = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Sagittal View', main_window)
    show_sagittal_view_action.setStatusTip('Show sagittal view')
    show_sagittal_view_action.triggered.connect(main_window.sagittal_viewer.show)
    menu.addAction(show_sagittal_view_action)

    #show_coronal_view_action = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Coronal View', main_window)
    #show_coronal_view_action.setStatusTip('Show coronal view')
    #show_coronal_view_action.triggered.connect(main_window.coronal_viewer.show)
    #menu.addAction(show_coronal_view_action)


def add_brush_menu(im_viewer, menu_bar):
    brush_menu = menu_bar.addMenu("Brushes")

    def add_brush(name, color_val, shortcut=None):
        color_action = QtWidgets.QAction(QtGui.QIcon(""), name, im_viewer)
        if shortcut:
            color_action.setShortcut(shortcut)
        brush_menu.addAction(color_action)
        color_action.triggered.connect(partial(im_viewer.set_color,
                                               color=QtGui.QColor(*color_val)))
        if im_viewer.brush_color is None:
            im_viewer.brush_color = QtGui.QColor(*color_val)
    # These don't need to be modified. Each class has background and foreground
    brushes = [
        ('Background', (0, 255, 0, 180), 'W'),
        ('Foreground', (255, 0, 0, 180), 'Q'),
        ('Eraser', (255, 205, 180, 0), 'E')
    ]
    for name, rgba, shortcut in brushes:
        add_brush(name, rgba, shortcut)


def add_class_menu(self, menu_bar):
    class_menu = menu_bar.addMenu('Classes')
    for i, class_name in enumerate(self.classes):
        class_action = QtWidgets.QAction(QtGui.QIcon('missing.png'), 
                                            class_name, self)
        # update class via nav to keep nav up to date whilst avoiding a double update.
        class_action.triggered.connect(partial(self.nav.cb.setCurrentIndex,
                                            self.classes.index(class_name)))
        class_action.setShortcut(str(i+1))
        class_menu.addAction(class_action)


def add_help_menu(self,  menu_bar):
    help_menu = menu_bar.addMenu('Help')
    license_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'License', self)
    license_btn.triggered.connect(self.show_license_window)
    help_menu.addAction(license_btn)

    about_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'About', self)
    about_btn.triggered.connect(self.show_about_window)
    help_menu.addAction(about_btn)
    
    shortcut_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Keyboard Shortcuts', self)
    shortcut_btn.triggered.connect(self.show_shortcut_window)
    help_menu.addAction(shortcut_btn)
    

def add_extras_menu(main_window, menu_bar, project_open=False):
    extras_menu = menu_bar.addMenu('Extras')

    if project_open:
        extend_dataset_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Extend dataset', main_window)
        def update_dataset_after_check():
            was_extended, file_names = check_extend_dataset(main_window,
                                                            main_window.dataset_dir,
                                                            main_window.image_fnames,
                                                            main_window.proj_file_path)
            if was_extended:
                main_window.image_fnames = file_names
                main_window.nav.all_fnames = file_names
                main_window.nav.update_nav_label()
        extend_dataset_btn.triggered.connect(update_dataset_after_check)
        extras_menu.addAction(extend_dataset_btn)


def check_extend_dataset(main_window, dataset_dir, prev_fnames, proj_file_path):

    all_image_names = [f for f in os.listdir(dataset_dir) if is_image(f)]

    new_image_names = [f for f in all_image_names if f not in prev_fnames]

    button_reply = QtWidgets.QMessageBox.question(main_window,
        'Confirm',
        f"There are {len(new_image_names)} new images in the dataset."
        " Are you sure you want to extend the project to include these new images?",
        QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, 
        QtWidgets.QMessageBox.No)

    if button_reply == QtWidgets.QMessageBox.Yes:
        # shuffle the new file names
        shuffle(new_image_names)
        # load the project json for reading and writing
        settings = json.load(open(proj_file_path, 'r'))
        # read the file_names
        all_file_names = settings['file_names'] + new_image_names
        settings['file_names'] = all_file_names

        # Add the new_files to the list
        # then save the json again
        json.dump(settings, open(proj_file_path, 'w'), indent=4)
        return True, all_file_names
    else:
        return False, all_image_names


def add_view_menu(window, im_viewer, menu_bar):
    """ Create view menu with options for
        * fit to view
        * actual size
        * toggle segmentation visibility
        * toggle annotation visibility
        * toggle image visibility
    """
    view_menu = menu_bar.addMenu('View')

    # Fit to view
    fit_to_view_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Fit to View', window)
    fit_to_view_btn.setShortcut('Ctrl+F')
    fit_to_view_btn.setStatusTip('Fit image to view')
    fit_to_view_btn.triggered.connect(im_viewer.graphics_view.fit_to_view)
    view_menu.addAction(fit_to_view_btn)

    # Actual size
    actual_size_view_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'), 'Actual size', window)
    actual_size_view_btn.setShortcut('Ctrl+A')
    actual_size_view_btn.setStatusTip('Show image at actual size')
    actual_size_view_btn.triggered.connect(im_viewer.graphics_view.show_actual_size)
    actual_size_view_btn.triggered.connect(im_viewer.graphics_view.show_actual_size)
    view_menu.addAction(actual_size_view_btn)

    # toggle segmentation visibility
    toggle_seg_visibility_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                  'Toggle segmentation visibility', window)
    toggle_seg_visibility_btn.setShortcut('S')
    toggle_seg_visibility_btn.setStatusTip('Show or hide segmentation')
    toggle_seg_visibility_btn.triggered.connect(im_viewer.show_hide_seg)
    view_menu.addAction(toggle_seg_visibility_btn)

    # toggle annotation visibility
    toggle_annot_visibility_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                    'Toggle annotation visibility', window)
    toggle_annot_visibility_btn.setShortcut('A')
    toggle_annot_visibility_btn.setStatusTip('Show or hide annotation')
    toggle_annot_visibility_btn.triggered.connect(im_viewer.show_hide_annot)
    view_menu.addAction(toggle_annot_visibility_btn)

    # toggle image visibility
    toggle_image_visibility_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                    'Toggle image visibility', window)
    toggle_image_visibility_btn.setShortcut('I')
    toggle_image_visibility_btn.setStatusTip('Show or hide image')
    toggle_image_visibility_btn.triggered.connect(im_viewer.show_hide_image)
    view_menu.addAction(toggle_image_visibility_btn)


    toggle_guide_image_visibility_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                    'Toggle guide image visibility', window)
    toggle_guide_image_visibility_btn.setShortcut('H')
    toggle_guide_image_visibility_btn.setStatusTip('Show or hide image')
    toggle_guide_image_visibility_btn.triggered.connect(im_viewer.show_hide_guide_image)
    view_menu.addAction(toggle_guide_image_visibility_btn)


    # toggle outline visibility
    toggle_outline_visibility_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                      'Toggle outline visibility', window)
    toggle_outline_visibility_btn.setShortcut('T')
    toggle_outline_visibility_btn.setStatusTip('Show or hide outline')
    toggle_outline_visibility_btn.triggered.connect(im_viewer.show_hide_outline)
    view_menu.addAction(toggle_outline_visibility_btn)

    up_slice_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                 'Move slice up', window)
    up_slice_btn.setShortcut('R')
    up_slice_btn.setStatusTip('move slice up')
    up_slice_btn.triggered.connect(im_viewer.slice_nav.up_slice)
    view_menu.addAction(up_slice_btn)

    down_slice_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                   'Move slice down', window)
    down_slice_btn.setShortcut('F')
    down_slice_btn.setStatusTip('move slice down')
    down_slice_btn.triggered.connect(im_viewer.slice_nav.down_slice)
    view_menu.addAction(down_slice_btn)

    # zoom in (V)
    zoom_in_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                'Zoom in', window)
    zoom_in_btn.setShortcut('V')
    zoom_in_btn.setStatusTip('Zoom in')
    zoom_in_btn.triggered.connect(im_viewer.zoom_in)
    view_menu.addAction(zoom_in_btn)

    # zoom in (V)
    zoom_out_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                                'Zoom out', window)
    zoom_out_btn.setShortcut('C')
    zoom_out_btn.setStatusTip('Zoom out')
    zoom_out_btn.triggered.connect(im_viewer.zoom_out)
    view_menu.addAction(zoom_out_btn)

    return view_menu



"""
Copyright (C) 2022 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from PyQt5 import QtWidgets
import os
import getpass

def create_lock_file(proj_dir, fname):
    """
    Create a lock file for the current user and the image they are annotating.
    """
    user_lock_dir = os.path.join(proj_dir, 'lock_files', getpass.getuser())
    if not os.path.isdir(user_lock_dir):
        os.makedirs(user_lock_dir)
    fpath = os.path.join(user_lock_dir, fname)
    f = open(fpath, 'w+')
    print('', file=f)
    f.close()

def delete_lock_files_for_current_user(proj_dir):
    """ get the current user
        and delete any lock files for this user
    """
    user_lock_dir = os.path.join(proj_dir, 'lock_files', getpass.getuser())
    if not os.path.isdir(user_lock_dir):
        os.makedirs(user_lock_dir)
    fnames = os.listdir(user_lock_dir)
    for f in fnames:
        os.remove(os.path.join(user_lock_dir, f))

def get_lock_file_path(proj_dir, fname):
    users = os.listdir(os.path.join(proj_dir,  'lock_files'))
    for u in users:
        fpath = os.path.join(proj_dir, 'lock_files', u, fname)
        if os.path.isfile(fpath):
            return u, fpath
    return False

def show_locked_message(proj_dir, fname):
    username = getpass.getuser()
    msg = QtWidgets.QMessageBox()
    uname, fpath = get_lock_file_path(proj_dir, fname)
    msg.setText(f"A lock file exists indicating"
                f' that {uname} is editing {fname}. \n'
                f" {fname} will be skipped and you will be shown the next image. \n"
                " If you wish to edit the file anyway, you can delete the "
                f"lock file located at {fpath}.")
    msg.show()
    return msg # need to return it so it doesn't go out of scope and disapear


"""
Show visibility status of segmentation, image and annotation.

Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=E0611, C0111, C0111, R0903, I1101
from PyQt5 import QtWidgets

class VisibilityWidget(QtWidgets.QWidget):

    def __init__(self, layout_class, parent, show_guide=False):
        super().__init__()
        self.initUI(layout_class, parent, show_guide)

    def initUI(self, layout_class, parent, show_guide):
        # container goes full width to allow contents to be center aligned within it.
        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        # left, top, right, bottom
        layout.setContentsMargins(0, 0, 0, 0)
        widget.setLayout(layout)

        container_layout = layout_class()
        self.setLayout(container_layout)
        container_layout.setContentsMargins(0, 0, 0, 0)

        seg_checkbox = QtWidgets.QCheckBox("Segmentation (S)")
        container_layout.addWidget(seg_checkbox)

        annot_checkbox = QtWidgets.QCheckBox("Annotation (A)")
        container_layout.addWidget(annot_checkbox)

        im_checkbox = QtWidgets.QCheckBox("Image (I)")
        container_layout.addWidget(im_checkbox)

        outline_checkbox = QtWidgets.QCheckBox("Outline (T)")
        container_layout.addWidget(outline_checkbox)

        # the guide image is optional and only enabled for some projects
        if show_guide:
            guide_image_checkbox = QtWidgets.QCheckBox("Guide (H)")
            container_layout.addWidget(guide_image_checkbox)
            guide_image_checkbox.setChecked(False)
            self.guide_image_checkbox = guide_image_checkbox
            self.guide_image_checkbox.stateChanged.connect(parent.guide_checkbox_change)

        seg_checkbox.setChecked(False)
        annot_checkbox.setChecked(True)
        im_checkbox.setChecked(True)
        outline_checkbox.setChecked(False)

  

        self.seg_checkbox = seg_checkbox
        self.annot_checkbox = annot_checkbox
        self.im_checkbox = im_checkbox
        self.outline_checkbox = outline_checkbox

        # connecting events to parent - tight coupling is OK if that's all we want.
        self.seg_checkbox.stateChanged.connect(parent.seg_checkbox_change)
        self.annot_checkbox.stateChanged.connect(parent.annot_checkbox_change)
        self.im_checkbox.stateChanged.connect(parent.im_checkbox_change)
        self.outline_checkbox.stateChanged.connect(parent.outline_checkbox_change)
 


""" Segment just one patch
Copyright (C) 2021 Abraham George Smith
Copyright (C) 2022 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import time
import torch
import socket
import ssl
import numpy as np
import zlib
import json
from pathlib import Path
import threading



import rp_annot as rpa
from skimage import img_as_float32
import numpy as np
from torch.nn.functional import softmax
from instructions import fix_config_paths

from startup import add_config_shape
import im_utils
from model_utils import load_model, get_latest_model_paths
from scp_utils import scp_transfer

cached_image = None
cached_image_fname = None
cached_model = None
cached_model_path = None
 

def start_server(sync_dir, ip, port):


    def create_server_socket(sync_dir, ip, port):
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        server_cert = os.path.join(Path.home(), 'root_painter_server.public_key')
        server_key = os.path.join(Path.home(), 'root_painter_server.private_key')
        client_certs = os.path.join(Path.home(), 'root_painter_client.public_key')

        context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        context.verify_mode = ssl.CERT_REQUIRED
        context.load_cert_chain(certfile=server_cert, keyfile=server_key)
        context.load_verify_locations(cafile=client_certs)

        # CUDA can be slow the first time it is used. We run this here so user interation is not
        # delayed by waiting for CUDA.
        print('starting CUDA..')
        _ = torch.from_numpy(np.zeros((100))).to(device)



        # Create a TCP/IP socket
        # Listen for incoming connections
        sock = socket.socket()
        print(f'Listening on {ip} port {port}')
        sock.bind((ip, port))
        sock.listen(1)
        while True:
            conn, client_address = sock.accept()
            conn = context.wrap_socket(conn, server_side=True)
            print("SSL established. Peer: {}".format(conn.getpeercert()))
            print('connection from', client_address)
            annot = None
            buf = b''
            while True:
                data = conn.recv(16)
                if data == b'end':
                    shape = np.frombuffer(buf[:16], dtype='int32')
                    annot1d = rpa.decompress(buf[16:], np.prod(shape))
                    annot = annot1d.reshape(shape)
                    buf = b''
                elif data == b'cfg':
                    segment_config = json.loads(buf.decode())
                    segment_config = fix_config_paths(sync_dir, segment_config)
                    segment_patch(segment_config, annot, conn)
                    buf = b''
                else:
                    buf += data
    

    server_thread = threading.Thread(target=create_server_socket, args=(sync_dir, ip, port))
    server_thread.start() # it doesn't finish.



def segment_patch(segment_config, annot_patch, conn):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    global cached_image
    global cached_image_fname
    global cached_model
    global cached_model_path

    start = time.time()
    fname = segment_config['file_name']
    dataset_dir = segment_config['dataset_dir']
    model_dir = segment_config['model_dir']
    classes = segment_config['classes']
    # Where do we take the patch from? (coord dont include padding)
    x_start = segment_config['x_start']
    y_start = segment_config['y_start']
    z_start = segment_config['z_start']
    x_end = segment_config['x_end']
    y_end = segment_config['y_end']
    z_end = segment_config['z_end']
    model_path = get_latest_model_paths(model_dir, 1)[0]
    fpath = os.path.join(dataset_dir, fname)

    if not os.path.isfile(fpath):
        print('Cannot load ', fpath, 'file does not exist')
        return

    # TODO: Get annot_dirs later - for now leave empty, as using single class only
    if fname == cached_image_fname:
        image = cached_image
    else:
        # load image for this patch
        (image, _, _, _) = im_utils.load_image_and_annot_for_seg(dataset_dir,
                                                                            [],
                                                                            fname)
        # loaded image is not longer pre-padded so padding must be assigned.
        image = im_utils.pad_3d(image, 17, 17, mode='reflect', constant_values=0)
        cached_image = image
        cached_image_fname = fname

    segment_config = add_config_shape(segment_config)
    # we know that the loaded image is padded so consider this when extracting the patch
    pad_z = (segment_config['in_d'] - segment_config['out_d']) // 2
    pad_y = (segment_config['in_w'] - segment_config['out_w']) // 2
    pad_x = pad_y

    # take the relevant region as specified in the instrction
    im_patch = image[
        z_start+pad_z:z_end+pad_z,
        y_start+pad_y:y_end+pad_y,
        x_start+pad_y:x_end+pad_x]

    # now normalise the tile (as this is done for all input to the network
    im_patch = im_utils.normalize_tile(img_as_float32(im_patch))
    model_input = torch.cuda.FloatTensor(1, 3, im_patch.shape[0], im_patch.shape[1], im_patch.shape[2])
    model_input[0, 0] = torch.from_numpy(im_patch.astype(np.float32)).to(device)

    bg_patch = annot_patch[0]
    fg_patch = annot_patch[1]

    model_input[0, 1] = torch.from_numpy(fg_patch).to(device)
    model_input[0, 2] = torch.from_numpy(bg_patch).to(device)

    # Use a cached model for improved performance
    if model_path == cached_model_path:
        cnn = cached_model
    else:
        cnn = load_model(model_path, classes).half()
        cached_model = cnn
        cached_model_path = model_path
    outputs = cnn(model_input.half())
    a = outputs.shape
    # bg channel index for each class in network output.
    class_idxs = [x * 2 for x in range(outputs.shape[1] // 2)]
    class_output_patches = [[] for _ in class_idxs]
    for i, class_idx in enumerate(class_idxs):
        # class_output = (batch_size, bg/fg, depth, height, width)
        class_output = outputs[:, class_idx:class_idx+2]
        softmaxed = softmax(class_output, 1) 
        foreground_probs = softmaxed[:, 1]  # just the foreground probability.
        predicted = foreground_probs > 0.5
        predicted = predicted.type(torch.cuda.ByteTensor)
        pred_np = predicted.data.detach().cpu().numpy()
        for out_tile in pred_np:
            class_output_patches[i].append(out_tile)
    # For now only the first class will be segmented 
    seg = class_output_patches[0][0]
    # send segmented region to client
    seg = seg.astype(bool)
    compressed = rpa.compress(seg.reshape(-1))
    compressed = zlib.compress(compressed)
    conn.sendall(compressed)
    conn.sendall(b'end')
    print('time to segment patch and transfer', time.time() - start)



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=C0111, W0511
# pylint: disable=E0401 # import error

import os
import warnings
import glob
import sys

import numpy as np
from skimage import color
from skimage.io import imread, imsave
from skimage import img_as_ubyte
from skimage.transform import resize
from skimage.measure import label
from skimage.morphology import binary_dilation, remove_small_holes
from PIL import Image
from PyQt5 import QtGui
import qimage2ndarray
import nibabel as nib
from scipy.ndimage import binary_fill_holes
from skimage.segmentation import flood
import nrrd
import SimpleITK as sitk
from pathlib import Path
from scipy.ndimage import zoom
import pandas as pd
import psycopg2

def is_image(fname):
    extensions = {
        '.npy', 'gz', 'nrrd', '.dicom',
        '.dcm', '.DCM', '.DICOM', '.sr', '.SR'
        }
    return any(fname.lower().endswith(ext) for ext in extensions)


def getDataFromDatabase(sql_config):
    connection = psycopg2.connect(
        host=sql_config['host'],
        database=sql_config['database'],
        user=sql_config['username'],
        password=sql_config['password'])
    sql = sql_config['query'].replace(
        "?table_name", "\"" + sql_config['table_name'] + "\"")
    sql = sql.replace(
        "?schema_name", "\"" + sql_config['schema_name'] + "\"")
    sql = sql.replace(
        "??", "\"")
    df = pd.read_sql_query(sql, connection)
    if len(df) == 0:
        print('The requested query does not have any data!')
    connection.close()

    return df['DcmPathFlatten'].tolist()


def get_recursive_files(input_path):
    filenames = []
    path = Path(input_path)
    for p in path.rglob("*"):
        if os.path.isdir(str(p)) is False:
            filenames.append(p._str)
    return filenames



def resizeVolume(img, output_size):
    factors = (output_size[0]/img.shape[1],
                output_size[1]/img.shape[2])

    return zoom(img, (1, factors[0], factors[1]))

def pad_image(image, pad_size):
    shap = image.shape
    if shap[0] < pad_size:
        to_pad = pad_size - shap[0]
        if (to_pad % 2) == 0:
            image = np.pad(image, ((to_pad/2,to_pad/2), (0,0), (0, 0)), 'constant')
        else:
            image = np.pad(image, ((int(to_pad/2)+1,int(to_pad/2)), (0,0), (0, 0)), 'constant')
    # else:
    #     image = image[0:pad_size, :, :]
    return image


def load_image(image_path):
    if image_path.endswith('.npy'):
        return np.load(image_path, mmap_mode='c')
    elif image_path.endswith('.nii.gz'):
        image = nib.load(image_path)
        image = np.array(image.dataobj)
    elif image_path.endswith('.nrrd'):
        image, header = nrrd.read(image_path)
    elif image_path.endswith(('.dicom', '.DICOM', '.dcm',
                              '.DCM',  '.sr', '.SR')):
        image = sitk.ReadImage(image_path)
        image = sitk.GetArrayFromImage(image)
       # print('image shape loaded (dcm)', image.shape)
    else:
        raise Exception(f"Unhandled file ending {image_path}")
    image = image.astype(np.int)
    return image

def load_annot(annot_path, img_data_shape):
    if annot_path.endswith(('.dicom', '.DICOM', '.dcm',
                            '.DCM',  '.sr', '.SR')):
        image = sitk.ReadImage(annot_path)
        annot_data = sitk.GetArrayFromImage(image)
        annot_data = np.array(annot_data, dtype=bool)
        print('annot data shape', annot_data.shape)
        
    else:
        annot_image = nib.load(annot_path)
        annot_data = np.array(annot_image.dataobj, dtype=bool)
  #      annot_data = np.rot90(annot_data, k=3)
      # print('annot data shape before permutation', annot_data.shape)
       # annot_data = np.moveaxis(annot_data, -1, 0) # depth moved to beginning
        print('annot data shape permutation', annot_data.shape)

    return annot_data


def load_seg(seg_path):
    if seg_path.endswith(('.dicom', '.DICOM', '.dcm',
                            '.DCM',  '.sr', '.SR')):
        image = sitk.ReadImage(seg_path)
        seg_data = sitk.GetArrayFromImage(image)
       # seg_data = resizeVolume(seg_data, (128, 128))
        seg_data = np.array(seg_data, dtype=bool)
        print('seg data shape', seg_data.shape)

        
    else:
        seg_image = nib.load(seg_path)
        seg_data = np.array(seg_image.dataobj, dtype=bool)
    #    seg_data = np.rot90(seg_data, k=3)
        print('seg_data shape before per', seg_data.shape)
       # seg_data = np.moveaxis(seg_data, -1, 0) # depth moved to beginning
        print('seg_data shape', seg_data.shape)
    # This issue may be related to file system issues.
    assert  len(seg_data.shape) == 3, f"seg shape is {seg_data.shape} for {seg_path}"
    return seg_data


def norm_slice(img, min_v, max_v, brightness_percent):
    if img.dtype != np.float32:
        img = img.astype(np.float32)
    bright_v = (brightness_percent / 100)
    img[img < min_v] = min_v
    img[img > max_v] = max_v
    img -= min_v
    img /= (max_v - min_v)
    img *= bright_v
    img[img > 1] = 1.0
    img *= 255
    return img


def annot_slice_to_pixmap(slice_np):
    """ convert slice from the numpy annotation data
        to a PyQt5 pixmap object """
    # for now fg and bg colors are hard coded.
    # later we plan to let the user specify these in the user interface.
    np_rgb = np.zeros((slice_np.shape[1], slice_np.shape[2], 4))
    np_rgb[:, :, 1] = slice_np[0] * 255 # green is bg
    np_rgb[:, :, 0] = slice_np[1] * 255 # red is fg
    np_rgb[:, :, 3] = np.sum(slice_np, axis=0) * 180 # alpha is defined
    q_image = qimage2ndarray.array2qimage(np_rgb)
    return QtGui.QPixmap.fromImage(q_image)


def get_outline_pixmap(seg_slice, annot_slice):
    seg_map = (seg_slice > 0).astype(np.int)
    annot_plus = (annot_slice[1] > 0).astype(np.int)
    annot_minus = (annot_slice[0] > 0).astype(np.int)

    # remove anything where seg is less than 0 as this is outside of the box
    seg_minus = (seg_slice < 0).astype(np.int)
    mask = ((((seg_map + annot_plus) - annot_minus) - seg_minus) > 0)
    dilated = binary_dilation(mask)
    outline = dilated.astype(np.int) - mask.astype(np.int)
    np_rgb = np.zeros((outline.shape[0], outline.shape[1], 4))
    np_rgb[outline > 0] = [255, 255, 0, 180]
    q_image = qimage2ndarray.array2qimage(np_rgb)
    return QtGui.QPixmap.fromImage(q_image)


def seg_slice_to_pixmap(slice_np):
    """ convert slice from the numpy segmentation data
        to a PyQt5 pixmap object """
    np_rgb = np.zeros((slice_np.shape[0], slice_np.shape[1], 4))
    np_rgb[slice_np > 0] = [0, 255, 255, 180]
    q_image = qimage2ndarray.array2qimage(np_rgb)
    return QtGui.QPixmap.fromImage(q_image)

def get_slice(volume, slice_idx, mode):
    if mode == 'axial':
        if len(volume.shape) > 3:
            slice_idx = (volume.shape[1] - slice_idx) - 1
            # if more than 3 presume first is channel dimension
            slice_data = volume[:, slice_idx, :, :]
        else:
            slice_idx = (volume.shape[0] - slice_idx) - 1
            slice_data = volume[slice_idx, :, :]
    elif mode == 'coronal':
        raise Exception("not yet implemented")
        if len(volume.shape) > 3:
            # if more than 3 presume first is channel dimension
            slice_data = volume[:, :, :, slice_idx]
        else:
            slice_data = volume[:, slice_idx, :]
    elif mode == 'sagittal':
        if len(volume.shape) > 3:
            # if more than 3 presume first is channel dimension
            slice_data = volume[:, :, :, slice_idx]
        else:
            slice_data = volume[:, :, slice_idx]
    else:
        raise Exception(f"Unhandled slice mode: {mode}")
    return slice_data


def store_annot_slice(annot_pixmap, annot_data, slice_idx, mode):
    """
    Update .annot_data at slice_idx
    so the values for fg and bg correspond to annot_pixmap)
    """
    slice_rgb_np = np.array(qimage2ndarray.rgb_view(annot_pixmap.toImage()))
    fg = slice_rgb_np[:, :, 0] > 0
    bg = slice_rgb_np[:, :, 1] > 0
    if mode == 'axial': 
        slice_idx = (annot_data.shape[1] - slice_idx) - 1
        annot_data[0, slice_idx] = bg
        annot_data[1, slice_idx] = fg
    elif mode == 'coronal':
        raise Exception("not yet implemented")
        annot_data[0, :, slice_idx, :] = bg
        annot_data[1, :, slice_idx, :] = fg
    elif mode == 'sagittal':
        #slice_idx = (annot_data.shape[3] - slice_idx) - 1
        annot_data[0, :, :, slice_idx] = bg
        annot_data[1, :, :, slice_idx] = fg
    else:
        raise Exception(f"Unhandled slice mode: {mode}")
    return annot_data


def get_num_regions(seg_data, annot_data):
    seg_map = (seg_data > 0).astype(np.int)
    annot_plus = (annot_data[1] > 0).astype(np.int)
    annot_minus = (annot_data[0] > 0).astype(np.int)
    # remove anything where seg is less than 0 as this is outside of the box
    corrected = (((seg_map + annot_plus) - annot_minus) > 0)
    labelled = label(corrected, connectivity=2)
    return len(np.unique(labelled)) - 1 # don't consider background a region.


def restrict_to_regions_containing_points(seg_data, annot_data, region_points):
    assert len(region_points), 'at least one region point must be specified'
    # restrict corrected structure to only the selected
    # connected region found at x,y,z
    # also remove small holes.
    seg_map = (seg_data > 0).astype(np.int)
    annot_plus = (annot_data[1] > 0).astype(np.int)
    annot_minus = (annot_data[0] > 0).astype(np.int)

    # remove anything where seg is less than 0 as this is outside of the box
    corrected = (((seg_map + annot_plus) - annot_minus) > 0)
    labelled = label(corrected, connectivity=2)
    holes_removed = 0
    removed_count = 0
    selected_component = None
    for x, y, z in region_points:
        selected_label = labelled[z, y, x]
        if selected_label == 0:
            error = "Selected region was background. Select a foreground region to keep."
            return annot_data, 0, 0, error
        if selected_component is None:
            selected_component = labelled == selected_label
        else:
            selected_component += labelled == selected_label
        labelled[selected_component] = -1

    disconnected_regions = labelled > 0
    # Then update the annotation so that these regions are now background.
    annot_data[0][disconnected_regions] = 1
    annot_data[1][disconnected_regions] = 0
    removed_count += len(np.unique(labelled[disconnected_regions]))
    # removing small holes
    # this was taking too long so we restrict to the object of interest.
    coords = np.where(selected_component > 0)
    min_z = np.min(coords[0])
    max_z = np.max(coords[0])
    min_y = np.min(coords[1])
    max_y = np.max(coords[1])
    min_x = np.min(coords[2])
    max_x = np.max(coords[2])
    # remove_small_holes
    roi_annot_plus = (annot_data[1, min_z: max_z, min_y:max_y, min_x:max_x] > 0).astype(np.int)
    roi_annot_minus = (annot_data[0,  min_z: max_z, min_y:max_y, min_x:max_x] > 0).astype(np.int)
    # remove anything where seg is less than 0 as this is outside of the box
    roi_corrected = (((seg_map[min_z: max_z, min_y:max_y, min_x:max_x] + roi_annot_plus) - roi_annot_minus) > 0)
    roi_corrected_no_holes = binary_fill_holes(roi_corrected).astype(np.int)
    roi_extra_fg = roi_corrected_no_holes - roi_corrected
    holes_removed += len(np.unique(label(roi_extra_fg))) - 1
    # Set the extra foreground from remove small holes to foreground in the annotation.
    annot_data[0, min_z: max_z, min_y:max_y, min_x:max_x][roi_extra_fg > 0] = 0
    annot_data[1, min_z: max_z, min_y:max_y, min_x:max_x][roi_extra_fg > 0] = 1
    return annot_data, removed_count, holes_removed, False


def fill_annot(annot_pixmap):
    # convert the current annot slice to numpy
    image = annot_pixmap.toImage()
    rgb_np = np.array(qimage2ndarray.rgb_view(image))
    fg_mask = rgb_np[:, :, 0]
    fg_mask = binary_fill_holes(fg_mask)
    fg_mask = fg_mask.astype(np.int)
    np_rgba = np.zeros((rgb_np.shape[0], rgb_np.shape[1], 4))
    # set fg annotation to be new fg mask
    # set bg annotation to be opposite of the fg mask
    np_rgba[:, :, :][fg_mask < 1] = [0, 255, 0, 180]
    np_rgba[:, :, :][fg_mask > 0] = [255, 0, 0, 180]

    q_image = qimage2ndarray.array2qimage(np_rgba)
    return QtGui.QPixmap.fromImage(q_image)



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=I1101,C0111,W0201,R0903,E0611, R0902, R0914
# too many statements
# pylint: disable=R0915
import sys
import os
from pathlib import PurePath
import json
from functools import partial
import traceback
from datetime import datetime
import time

import numpy as np
from scipy import ndimage
from scipy.ndimage import binary_fill_holes
import nibabel as nib

from skimage.io import use_plugin
from PyQt5 import QtWidgets
from PyQt5 import QtGui
from PyQt5 import QtCore
from PyQt5.QtCore import Qt

from view_state import ViewState
from about import AboutWindow, LicenseWindow, ShortcutWindow
from create_project import CreateProjectWidget
from im_viewer import ImViewer, ImViewerWindow
from nav import NavWidget
from file_utils import penultimate_fname_with_segmentation
from file_utils import get_annot_path
from file_utils import maybe_save_annotation_3d
from instructions import send_instruction
from contrast_slider import ContrastSlider
import im_utils
import menus
from segment import segment_full_image
from lock import create_lock_file, delete_lock_files_for_current_user, get_lock_file_path, show_locked_message
from pathlib import Path
from im_utils import getDataFromDatabase


use_plugin("pil")

class RootPainter(QtWidgets.QMainWindow):

    closed = QtCore.pyqtSignal()

    def __init__(self, sync_dir, contrast_presets, config, server_ip=None, server_port=None):
        super().__init__()
        self.sync_dir = sync_dir
        self.config = config
        self.instruction_dir = sync_dir / 'instructions'
        self.send_instruction = partial(send_instruction,
                                        instruction_dir=self.instruction_dir,
                                        sync_dir=sync_dir)
        self.contrast_presets = contrast_presets
        self.view_state = ViewState.LOADING_SEG
        self.auto_complete_enabled = server_ip and server_port # aka patch_update
        # for scp communication from server to client.
        self.server_ip = server_ip
        self.server_port = server_port

        self.tracking = False
        self.seg_mtime = None
        self.im_width = None
        self.im_height = None
        self.annot_data = None
        self.seg_data = None
        

        # for patch segment, useful for knowing how much annotation to send to the server.
        self.input_shape = (52, 228, 228)
        self.output_shape = (18, 194, 194)

        self.lines_to_log = []
        self.log_debounce = QtCore.QTimer()
        self.log_debounce.setInterval(500)
        self.log_debounce.setSingleShot(True)
        self.log_debounce.timeout.connect(self.log_debounced)
        self.settings = self.get_config()
        self.initUI()

    def get_config(self):
        settings_path = os.path.join(Path.home(), 'root_painter_settings.json')
        settings = json.load(open(settings_path, 'r'))
        return settings

    def initUI(self):
        if len(sys.argv) < 2:
            self.init_missing_project_ui()
            return

        fname = sys.argv[1]
        if os.path.splitext(fname)[1] == '.seg_proj':
            proj_file_path = os.path.abspath(sys.argv[1])
            self.open_project(proj_file_path)
        else:
            # only warn if -psn not in the args. -psn is in the args when
            # user opened app in a normal way by clicking on the Application icon.
            if not '-psn' in sys.argv[1]:
                QtWidgets.QMessageBox.about(
                    self,
                    'Error',
                    f"{sys.argv[1]} ' is not a valid "
                    "segmentation project (.seg_proj) file")
            self.init_missing_project_ui()

    def get_train_annot_dir(self):
        # taking into account the current class.
        if len(self.classes) > 1:
            return self.proj_location / 'annotations' / self.cur_class / 'train'
        return self.proj_location / 'annotations' / 'train'

    def get_val_annot_dir(self):
        # taking into account the current class.
        if len(self.classes) > 1:
            return self.proj_location / 'annotations' / self.cur_class / 'val'
        return self.proj_location / 'annotations' / 'val'

    def open_project(self, proj_file_path):
        # extract json
        with open(proj_file_path, 'r') as json_file:
            settings = json.load(json_file)
            self.dataset_dir = PurePath(settings['dataset'])
            
            if 'guide_image_dir' in settings:
                self.guide_image_dir = self.sync_dir / 'datasets' / PurePath(settings['guide_image_dir'])

            self.proj_location = self.sync_dir / PurePath(settings['location'])
          #  self.image_fnames = settings['file_names']
            self.image_fnames = getDataFromDatabase(self.config)
            self.seg_dir = self.proj_location / 'segmentations'
            self.log_dir = self.proj_location / 'logs'

            self.train_seg_dirs = []
            self.train_annot_dirs = []
            self.val_annot_dirs = []
            # if going with a single class or old style settings
            # then use old style project structure with single train and val
            # folder, without the class name being specified
            if "classes" in settings and len(settings['classes']) > 1:
                # if more than one class is present then create train and val folders for each class
                self.classes = settings['classes']
                self.cur_class = self.classes[0]
                for c in self.classes:
                    self.train_annot_dirs.append(self.proj_location / 'annotations' / c / 'train')
                    self.val_annot_dirs.append(self.proj_location / 'annotations' / c / 'val')
                    self.train_seg_dirs.append(self.proj_location / 'train_segmentations' / c )
            else:         
                self.classes = ['annotations'] # default class for single class project.
                self.cur_class = self.classes[0]
                self.train_annot_dirs = [self.proj_location / 'annotations' / 'train']
                self.val_annot_dirs = [self.proj_location / 'annotations' / 'val']
                self.train_seg_dirs = [self.proj_location / 'train_segmentations']

            self.model_dir = self.proj_location / 'models'
            self.message_dir = self.proj_location / 'messages'

            # If there are any segmentations which have already been saved
            # then go through the segmentations in the order specified
            # by self.image_fnames
            # and set fname (current image) to be the penultimate image with a segmentation
            # The client will always segment the image after the one being viewed. So we don't
            # show the last image segmented until the user navigates to it, or we will constantly
            # move forwards through the dataset without annotating images 
            # (if we simply close and re-open the client)
            last_with_seg = penultimate_fname_with_segmentation(self.image_fnames, self.seg_dir)
            if last_with_seg:
                fname = last_with_seg
            else:
                fname = self.image_fnames[0]

            # set first image from project to be current image
            self.image_path = os.path.join(self.dataset_dir, fname)
            self.update_window_title()
            self.annot_path = get_annot_path(fname, self.get_train_annot_dir(),
                                             self.get_val_annot_dir())
            self.init_active_project_ui()

            self.track_changes()

    def log_debounced(self):
        """ write to log file only so often to avoid lag """
        with open(os.path.join(self.log_dir, 'client.csv'), 'a+') as log_file:
            while self.lines_to_log:
                line = self.lines_to_log[0]
                log_file.write(line)
                self.lines_to_log = self.lines_to_log[1:]

    def log(self, message):
        self.lines_to_log.append(f"{datetime.now()}|{time.time()}|{message}\n")
        self.log_debounce.start() # write after 1 second

    def update_file(self, fpath):
        """ Invoked when the file to view has been changed by the user.
            Show image file and it's associated annotation and segmentation """
        # save annotation for current file before changing to new file.
        self.log(f'update_file_start,fname:{os.path.basename(fpath)},view_state:{self.view_state}')

        delete_lock_files_for_current_user(self.proj_location) 
        lock_file_path = get_lock_file_path(self.proj_location, os.path.basename(fpath))
        if lock_file_path:
            self.msg = show_locked_message(self.proj_location, os.path.basename(fpath))
            # if a file is locked then show a warning to the user
            self.nav.update_to_next_image()
            return

        create_lock_file(self.proj_location, os.path.basename(fpath)) 
        self.tracking = False # take a break from tracking until we get the next image.
        
        if self.view_state == ViewState.ANNOTATING:
            self.save_annotation()
        #self.fname = os.path.basename(fpath)
        # basename = os.path.basename(fpath)
        # dirname = os.path.basename(os.path.dirname(fpath))
        # self.fname = os.path.join(dirname, basename)
        _, _, self.fname = fpath.partition(
            os.path.basename(self.dataset_dir._str) + os.path.sep)
        self.image_path = os.path.join(self.dataset_dir._str, self.fname)
        #etDataFromDatabase(self.settings)
        self.img_data = im_utils.load_image(self.image_path) 
        # if a guide image directory is specified - TODO: Consider removing guide image functionality if it isn't used frequently
        if hasattr(self, 'guide_image_dir'):
            guide_image_path = os.path.join(os.path.join(self.guide_image_dir, self.fname))
            # and a guide image is available for the current image.
            if os.path.isfile(guide_image_path):
                self.guide_img_data = im_utils.load_image(guide_image_path)
            else:
                pass
                # no guide image found for guide_image_path - it's optional anyway
        else:
            pass
            # no guide image directory found - it's optional anyway

        self.update_annot_and_seg()

        self.contrast_slider.update_range(self.img_data)
        self.update_window_title()

        # only segment if a segmentation is missing.
        if not os.path.isfile(self.get_seg_path()):
            print('no segmentation found at', self.get_seg_path())
            segment_full_image(self, self.fname) # segment current image.
        
        # segment the next image also.
        
        cur_im_idx = self.image_fnames.index(self.fname)
        if cur_im_idx < len(self.image_fnames):    
            next_im_fname = self.image_fnames[cur_im_idx+1]
            num_segmentations = len(self.get_all_seg_paths())
            # dont segment the next image early on in training (first 10 images).
            # We don't want the user to have to correct segmentations from old models at this point in training.
            if num_segmentations > 10 and not os.path.isfile(self.get_seg_path(next_im_fname)):
                segment_full_image(self, next_im_fname) 


        self.log(f'update_file_end,fname:{os.path.basename(fpath)},view_state:{self.view_state}')

    def navigate_to_top_of_structure(self, roi_connected_structure):
        # Used to speed things up when we first open an image, as this is where a user 
        # should start making the corrections.
        for i in range(roi_connected_structure.shape[0]):
            if np.any(roi_connected_structure[i]):
                slice_idx = (roi_connected_structure.shape[0] - i) - 1
                self.axial_viewer.slice_nav.slider.setValue(slice_idx)
                self.axial_viewer.slice_nav.value_changed()
                break


    def update_annot_and_seg(self):
        self.annot_path = None
        self.view_state = ViewState.LOADING_SEG 
        if self.fname:
            self.annot_path = get_annot_path(self.fname,
                                             self.get_train_annot_dir(),
                                             self.get_val_annot_dir())
        
        if self.annot_path and os.path.isfile(self.annot_path):
            self.annot_data = im_utils.load_annot(self.annot_path, self.img_data.shape)
        else:
            # otherwise create empty annotation array
            # if we are working with 3D data (npy file) and the
            # file hasn't been found then create an empty array to be
            # used for storing the annotation information.
            # channel for bg (0) and fg (1)
            self.annot_data = np.zeros([2] + list(self.img_data.shape))

        if self.fname and os.path.isfile(self.get_seg_path()):
            print('load seg')
            self.log(f'load_seg,fname:{os.path.basename(self.get_seg_path())}')
            self.seg_data = im_utils.load_seg(self.get_seg_path())
            self.view_state = ViewState.ANNOTATING
            self.update_segmentation()
        else:
            # it should come later
            self.seg_data = None
           
        for v in self.viewers:
            v.update_image()
            v.update_cursor()
            # hide the segmentation if we don't have it
            if self.seg_data is None and v.seg_visible:
                # show seg in order to show the loading message
                v.show_hide_seg()

        """
        if self.seg_data is not None:
            # assign bg annotations to regions that are not the largest
            import cc3d
            if not np.any(self.annot_data):
                print('run cc3d')
                labels_out = cc3d.connected_components(self.seg_data) # 26-connected
                voxel_counts = cc3d.statistics(labels_out)['voxel_counts']
                to_keep_idx = voxel_counts[1:].argmax()
                if len(voxel_counts) < 30:
                    for label_idx in range(len(voxel_counts)):
                        print(label_idx, 'of', len(voxel_counts))
                        if label_idx == to_keep_idx:
                            biggest_region_mask = labels_out == (to_keep_idx + 1)
                            roi_corrected_no_holes = binary_fill_holes(biggest_region_mask).astype(np.int)
                            roi_extra_fg = roi_corrected_no_holes - biggest_region_mask
                            self.annot_data[1][roi_extra_fg > 0] = 1 # set fg regions to remove holes.
                        else:
                            fg_region_label = label_idx + 1
                            mask = labels_out == fg_region_label
                            self.annot_data[0][mask] = 1 # set bg regions to remove extra components
                    self.update_viewer_annot_slice() # update view so next view change doesnt update the annot with previously displayed annot.
                    print('navigate to top')
                    self.navigate_to_top_of_structure(roi_corrected_no_holes)
                    print('done navigating to top')
                else:
                    print('Did not automatically correct disconnected regions as there are over 30') # it takes too long.
            else:
                print('found annot so not automatically remoing small holes and regions')
        """

    def update_class(self, class_name):
        # Save current annotation (if it exists) before moving on
        self.save_annotation()
        self.cur_class = class_name
        self.annot_path = get_annot_path(self.fname,
                                         self.get_train_annot_dir(),
                                         self.get_val_annot_dir())
        for v in self.viewers:
            v.scene.history = []
            v.scene.redo_list = []
        self.update_annot_and_seg()

    def get_seg_path(self, fname=None):
        if fname is None:
            fname = self.fname
        if fname.endswith(
            ('.dcm', '.dicom', '.sr', '.DCM', '.DICOM', '.SR')):
            seg_fname = fname.replace('.dcm', '.nii.gz').replace('.DCM', '.nii.gz')
        else:
            seg_fname = fname.replace('.nrrd', '.nii.gz')
        #seg_fname = fname.replace('.nrrd', '.nii.gz')
        # just seg path for current class.
        if hasattr(self, 'classes') and len(self.classes) > 1:
            return os.path.join(self.seg_dir,
                                self.cur_class,
                                seg_fname)
        return os.path.join(self.seg_dir, seg_fname)


    def get_train_seg_path(self, fname=None):
        if fname is None:
            fname = self.fname
        if fname.endswith(
            ('.dcm', '.dicom', '.sr', '.DCM', '.DICOM', '.SR')):
            seg_fname = fname.replace('.dcm', '.nii.gz').replace('.DCM', '.nii.gz')
        else:
            seg_fname = fname.replace('.nrrd', '.nii.gz')
        #seg_fname = fname.replace('.nrrd', '.nii.gz')
        if hasattr(self, 'classes') and len(self.classes) > 1:
            return os.path.join(self.proj_location,
                                'train_segmentations',
                                self.cur_class,
                                seg_fname)
        return os.path.join(self.proj_location, 'train_segmentations', seg_fname)


    def get_all_seg_paths(self):
        if hasattr(self, 'classes') and len(self.classes) > 1:
            spaths = []
            for c in self.classes:
                spaths.append(os.path.join(self.seg_dir,
                                self.cur_class,
                                self.fname))
            return spaths
        return [os.path.join(self.seg_dir, self.fname)]

    def update_segmentation(self):
        # if seg file is present then load.
        if os.path.isfile(self.get_seg_path()):
            self.seg_mtime = os.path.getmtime(self.get_seg_path())
            self.nav.next_image_button.setText('Save && Next >')
            self.nav.next_image_button.setEnabled(True)
        else:
            self.seg_mtime = None
            self.nav.next_image_button.setEnabled(False)
            self.nav.next_image_button.setText('Loading Segmentation...')

    def set_seg_loading(self):
        """ Transition to loading segmentation """
        self.seg_mtime = None
        self.nav.next_image_button.setEnabled(False)
        self.nav.next_image_button.setText('Loading Segmentation...')
        self.view_state = ViewState.LOADING_SEG
        for v in self.viewers:
            v.scene.last_x = None
            v.scene.last_y = None
            if not v.seg_visible:
                # show seg in order to show the loading message
                v.show_hide_seg()

    def show_open_project_widget(self):
        options = QtWidgets.QFileDialog.Options()
        default_loc = self.sync_dir / 'projects'
        file_path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Load project file",
            str(default_loc),
            "Segmentation project file (*.seg_proj)",
            options=options)

        if file_path:
            self.open_project(file_path)

    def show_create_project_widget(self):
        print("Open the create project widget..")
        self.create_project_widget = CreateProjectWidget(self.sync_dir)
        self.create_project_widget.show()
        self.create_project_widget.created.connect(self.open_project)

    def init_missing_project_ui(self):
        ## Create project menu
        # project has not yet been selected or created
        # need to open minimal interface which allows users
        # to open or create a project.

        menu_bar = self.menuBar()
        self.menu_bar = menu_bar
        self.menu_bar.clear()
        self.project_menu = menu_bar.addMenu("Project")

        # Open project
        self.open_project_action = QtWidgets.QAction(QtGui.QIcon(""), "Open project", self)
        self.open_project_action.setShortcut("Ctrl+O")

        self.project_menu.addAction(self.open_project_action)
        self.open_project_action.triggered.connect(self.show_open_project_widget)

        # Create project
        self.create_project_action = QtWidgets.QAction(QtGui.QIcon(""), "Create project", self)
        self.create_project_action.setShortcut("Ctrl+C")
        self.project_menu.addAction(self.create_project_action)
        self.create_project_action.triggered.connect(self.show_create_project_widget)

        menus.add_help_menu(self, menu_bar)
        menus.add_extras_menu(self, menu_bar)

        # Add project btns to open window (so it shows something useful)
        project_btn_widget = QtWidgets.QWidget()
        self.setCentralWidget(project_btn_widget)

        layout = QtWidgets.QHBoxLayout()
        project_btn_widget.setLayout(layout)
        open_project_btn = QtWidgets.QPushButton('Open existing project')
        open_project_btn.clicked.connect(self.show_open_project_widget)
        layout.addWidget(open_project_btn)

        create_project_btn = QtWidgets.QPushButton('Create new project')
        create_project_btn.clicked.connect(self.show_create_project_widget)
        layout.addWidget(create_project_btn)

        self.setWindowTitle("RootPainter3D - Not approved for clinical use.")
        self.resize(layout.sizeHint())


    def show_license_window(self):
        self.license_window = LicenseWindow()
        self.license_window.show()

    def show_about_window(self):
        self.about_window = AboutWindow()
        self.about_window.show()

    def show_shortcut_window(self):
        self.shortcut_window = ShortcutWindow()
        self.shortcut_window.show()

    def update_window_title(self):
        proj_dirname = os.path.basename(self.proj_location)
        self.setWindowTitle(f"RootPainter3D {proj_dirname}"
                            f" {os.path.basename(self.image_path)}"
                            " - Not approved for clinical use")

    def closeEvent(self, _):
        if hasattr(self, 'proj_location'):
            delete_lock_files_for_current_user(self.proj_location)
        if hasattr(self, 'contrast_slider'):
            self.contrast_slider.close()
        if hasattr(self, 'sagittal_viewer'):
            self.sagittal_viewer.close()
        if hasattr(self, 'coronal_viewer'):
            self.coronal_viewer.close()


    def contrast_updated(self):
        self.update_viewer_image_slice()
        self.update_viewer_guide()


    def update_viewer_image_slice(self):
        for v in self.viewers:
            if v.isVisible():
                v.update_image_slice()

    def update_viewer_annot_slice(self):
        for v in self.viewers:
            if v.isVisible():
                v.update_annot_slice()

    def update_viewer_outline(self):
        for v in self.viewers:
            if v.isVisible():
                v.update_outline()

    def update_viewer_guide(self):
        for v in self.viewers:
            if v.isVisible():
                v.update_guide_image_slice()

    def before_nav_change(self):
        """
        I'm trying to make sure the user doesn't forget to remove
        disconnected regions.
        """
        if self.seg_data is None:
            button_reply = QtWidgets.QMessageBox.question(
                self,
                'Confirm',
                f"The segmentation has not yet loaded for this image. "
                "Are you sure you want to proceed to the next image?",
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, 
                QtWidgets.QMessageBox.No)
            return button_reply == QtWidgets.QMessageBox.Yes

        # return False to block nav change
        num_regions = im_utils.get_num_regions(self.seg_data,
                                               self.annot_data)
        if num_regions == 1:
            return True
        button_reply = QtWidgets.QMessageBox.question(
            self,
            'Confirm',
            f"There are {num_regions} regions in this image. "
            "Are you sure you want to proceed to the next image?",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, 
            QtWidgets.QMessageBox.No)

        return button_reply == QtWidgets.QMessageBox.Yes

    def init_active_project_ui(self):
        # container for both nav and im_viewer.
        container = QtWidgets.QWidget()
        container_layout = QtWidgets.QVBoxLayout()
        container_layout.setContentsMargins(0, 0, 0, 0)
        self.container = container
        self.container_layout = container_layout
        container.setLayout(container_layout)
        self.setCentralWidget(container)

        self.viewers_container = QtWidgets.QWidget()
        self.viewers_layout = QtWidgets.QHBoxLayout()
        self.viewers_container.setLayout(self.viewers_layout)

        self.axial_viewer = ImViewer(self, 'axial')
        self.sagittal_viewer = ImViewerWindow(self, 'sagittal')
        self.sagittal_viewer.show()
        
        #self.coronal_viewer = ImViewerWindow(self, 'coronal')
        self.viewers = [self.axial_viewer, self.sagittal_viewer] #, self.coronal_viewer]
        self.viewers_layout.addWidget(self.axial_viewer)

        container_layout.addWidget(self.viewers_container)
        self.contrast_slider = ContrastSlider(self.contrast_presets)
        self.contrast_slider.changed.connect(self.contrast_updated)

        self.nav = NavWidget(self.image_fnames,
                             self.classes,
                             self.before_nav_change)

        # bottom bar right
        bottom_bar_r = QtWidgets.QWidget()
        bottom_bar_r_layout = QtWidgets.QVBoxLayout()
        bottom_bar_r.setLayout(bottom_bar_r_layout)
        self.axial_viewer.bottom_bar_layout.addWidget(bottom_bar_r)
        self.hu_label = QtWidgets.QLabel()
        self.hu_label.setAlignment(Qt.AlignRight)
        bottom_bar_r_layout.addWidget(self.hu_label)

        
        # Nav
        self.nav.file_change.connect(self.update_file)
        self.nav.class_change.connect(self.update_class)
        self.nav.image_path = self.image_path
        self.nav.update_nav_label()

        # info label
        info_container = QtWidgets.QWidget()
        info_container_layout = QtWidgets.QHBoxLayout()
        info_container_layout.setAlignment(Qt.AlignCenter)
        info_label = QtWidgets.QLabel()
        info_label.setText("")
        info_container_layout.addWidget(info_label)
        # left, top, right, bottom
        info_container_layout.setContentsMargins(0, 0, 0, 0)
        info_container.setLayout(info_container_layout)
        self.info_label = info_label
        # add nav and info label to the axial viewer.
        bottom_bar_r_layout.addWidget(info_container)
        bottom_bar_r_layout.addWidget(self.nav)

        self.add_menu()
        self.resize(container_layout.sizeHint())
        self.axial_viewer.update_cursor()
        self.update_file(self.image_path)

        def view_fix():
            """ started as hack for linux bug.
                now used for setting defaults """
            self.axial_viewer.update_cursor()
            # These are causing issues on windows so commented out until I find a better solution
            # self.set_to_left_half_screen()
            # self.sagittal_viewer.set_to_right_half_screen()
            self.set_default_view_size()

        QtCore.QTimer.singleShot(100, view_fix)


    def set_default_view_size(self):
        # sensible defaults for CT scans
        self.axial_viewer.graphics_view.zoom = 2.4
        self.sagittal_viewer.graphics_view.zoom = 2.0
        self.axial_viewer.graphics_view.update_zoom()
        self.sagittal_viewer.graphics_view.update_zoom()

    def set_to_left_half_screen(self): 
        screen_shape = QtWidgets.QDesktopWidget().screenGeometry()
        w = screen_shape.width() // 2
        h = screen_shape.height()
        x = 0
        y = 0
        self.setGeometry(x, y, w, h)

    def track_changes(self):
        if self.tracking:
            return
        self.tracking = True
        def check():
            # check for any messages
            messages = os.listdir(str(self.message_dir))

            for m in messages:
                if hasattr(self, 'info_label'):
                    self.info_label.setText(m)
                try:
                    # Added try catch because this error happened (very rarely)
                    # PermissionError: [WinError 32]
                    # The process cannot access the file because `it is
                    # being used by another process
                    os.remove(os.path.join(self.message_dir, m))
                except Exception as e:
                    print('Caught exception when trying to detele msg', e)
                                
            # if a segmentation exists (on disk)
            if self.fname and os.path.isfile(self.get_seg_path()):
                try:
                    # seg mtime is not actually used any more.
                    new_mtime = os.path.getmtime(self.get_seg_path())

                    # seg_mtime is None before the seg is loaded.
                    if self.seg_mtime is None or new_mtime != self.seg_mtime:
                        print('load new seg now')
                        self.log(f'load_seg,fname:{os.path.basename(self.get_seg_path())}')
                        self.seg_data = im_utils.load_seg(self.get_seg_path())
                        self.axial_viewer.update_seg_slice()
                        # Change to annotating state.                        
                        self.view_state = ViewState.ANNOTATING
                        for v in self.viewers:
                            v.update_cursor()
                            # for some reason cursor doesn't update straight away sometimes.
                            # trigger again half a second later to make sure 
                            # the correct cursor is shown.
                            QtCore.QTimer.singleShot(500, v.update_cursor)
                            if v.isVisible():
                                v.update_seg_slice()

                        self.seg_mtime = new_mtime
                        self.nav.next_image_button.setText('Save && Next >')
                        self.nav.next_image_button.setEnabled(True)
                    else:
                        pass
           
                except Exception as e:
                    print(f'Exception loading segmentation,{e},{traceback.format_exc()}')
                    # sometimes problems reading file.
                    # don't worry about this exception
            else:
                pass
            QtCore.QTimer.singleShot(500, check)
        QtCore.QTimer.singleShot(500, check)

    def close_project_window(self):
        self.close()
        self.closed.emit()

    def add_menu(self):
        menu_bar = self.menuBar()
        menu_bar.clear()

        self.project_menu = menu_bar.addMenu("Project")
        # Open project
        self.close_project_action = QtWidgets.QAction(QtGui.QIcon(""), "Close project", self)
        self.project_menu.addAction(self.close_project_action)
        self.close_project_action.triggered.connect(self.close_project_window)
        menus.add_edit_menu(self, self.axial_viewer, menu_bar)


        #options_menu = menu_bar.addMenu("Options")

        self.menu_bar = menu_bar

        # add brushes menu for axial slice navigation
        menus.add_brush_menu(self.axial_viewer, self.menu_bar)

        # add view menu for axial slice navigation.
        view_menu = menus.add_view_menu(self, self.axial_viewer, self.menu_bar)
        self.add_contrast_setting_options(view_menu)

        menus.add_network_menu(self, self.menu_bar)
        menus.add_windows_menu(self)
        if len(self.classes) > 1:
            menus.add_class_menu(self, self.menu_bar)
        menus.add_help_menu(self, self.menu_bar)
        menus.add_extras_menu(self, menu_bar, project_open=True)


    def add_contrast_setting_options(self, view_menu):
        preset_count = 0
        for preset in self.contrast_presets:
            def add_preset_option(new_preset, preset_count):
                preset = new_preset
                preset_btn = QtWidgets.QAction(QtGui.QIcon('missing.png'),
                                               f'{preset} contrast settings', self)
                preset_btn.setShortcut(QtGui.QKeySequence(f"Alt+{preset_count}"))
                preset_btn.setStatusTip(f'Use {preset} contrast settings')
                def on_select():
                    self.contrast_slider.preset_selected(preset)
                preset_btn.triggered.connect(on_select)
                view_menu.addAction(preset_btn)
            preset_count += 1
            add_preset_option(preset, preset_count)

    def stop_training(self):
        self.info_label.setText("Stopping training...")
        content = {"message_dir": self.message_dir}
        self.send_instruction('stop_training', content)

    def start_training(self):
        self.info_label.setText("Starting training...")
        # 3D just uses the name of the first class
        content = {
            "model_dir": self.model_dir,
            "dataset_dir": self.dataset_dir,
            "train_annot_dirs": self.train_annot_dirs,
            "train_seg_dirs": self.train_seg_dirs,
            "val_annot_dirs": self.val_annot_dirs,
            "seg_dir": self.seg_dir,
            "log_dir": self.log_dir,
            "message_dir": self.message_dir,
            "classes": self.classes
        }
        self.send_instruction('start_training', content)

    def save_annotation(self):
        if self.annot_data is not None:
            for v in self.viewers:
                v.store_annot_slice()
           # _, _, fname = self.get_seg_path().partition('segmentations/')
            _, _, fname = self.get_seg_path().partition('segmentations/')
        #    fname = os.path.basename(self.get_seg_path())
            self.annot_path = maybe_save_annotation_3d(self.img_data.shape,
                                                       self.annot_data,
                                                       self.annot_path,
                                                       fname,
                                                       self.get_train_annot_dir(),
                                                       self.get_val_annot_dir(),
                                                       self.log)
            if self.annot_path:
                #if self.auto_complete_enabled:
                # also save the segmentation, as this updated due to patch updates (potencially).
                #print('seg_data shape before permute', self.seg_data.shape)
                #seg_data_repermute = np.moveaxis(self.seg_data, -1, 0) 
                #print('seg_data shape before permute',seg_data_repermute.shape)
                
                out_path_seg = self.get_train_seg_path()
                dir_to_create = os.path.dirname(out_path_seg)
                Path(dir_to_create).mkdir(parents=True, exist_ok=True)
                    
                img = nib.Nifti1Image(self.seg_data.astype(np.int8), np.eye(4))
                img.to_filename(out_path_seg)
                # if annotation was saved to train 
                if str(self.get_train_annot_dir()) in self.annot_path:
                    # if there are at least 4 train annotations (happens after
                    # saving annotaiton for image 5) then save segmentation for
                    # training. The first 4 segmentations are not fully
                    # corrected per the new protocol from image 5 and onwards
                    # all segmentations will be saved for training as they are
                    # fully corrected.
                    

                    if len(os.listdir(self.get_train_annot_dir())) >= 4:
                        img.to_filename(out_path_seg)
                else:
                    # otherwise if it was saved to validation then start training
                    # as we now believe there is training and validation data.
                    self.start_training()



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

#pylint: disable=I1101,C0111,W0201,R0903,E0611,R0902,R0914,W0703
import sys
import os
from pathlib import Path
import json
import traceback

from PyQt5 import QtWidgets
from root_painter import RootPainter
import yaml
import argparse


def init_root_painter(config):
    sys.argv = [sys.argv[0]]
    app = QtWidgets.QApplication(sys.argv)
    settings_path = os.path.join(Path.home(), 'root_painter_settings.json')
    try:
        # if the settings file does not exist then create it with
        # a user specified sync_dir
        if not os.path.isfile(settings_path):
            msg = QtWidgets.QMessageBox()
            output = "Sync directory not specified. Please specify a sync directory."
            msg.setText(output)
            msg.exec_()
            dir_path = QtWidgets.QFileDialog.getExistingDirectory()
            if not dir_path:
                exit()
                
            with open(settings_path, 'w') as json_file:
                content = {
                    "sync_dir": os.path.abspath(dir_path),
                    "contrast_presets": {
                        'Mediastinal': [-125, 250, 100]
                    }
                }
                json.dump(content, json_file, indent=4)

        settings = json.load(open(settings_path, 'r'))

        sync_dir = Path(settings['sync_dir'])
        contrast_presets = settings['contrast_presets']
        server_ip = None
        server_port = None
        if "auto_complete" in settings and settings['auto_complete'] == True:    
            server_ip = settings["server_ip"]
            server_port = settings["server_port"]
        def reopen():
            main_window = RootPainter(sync_dir, contrast_presets, config, server_ip, server_port)
            main_window.closed.connect(reopen)
            main_window.show()

        main_window = RootPainter(sync_dir, contrast_presets, config, server_ip, server_port)
        # close project causes reopen with missing project UI
        main_window.closed.connect(reopen)
        main_window.show()
        sys.exit(app.exec_())

    except Exception as e:
        msg = QtWidgets.QMessageBox()
        output = f"""
        repr(e): {repr(e)}
        traceback.format_exc(): {traceback.format_exc()}
        sys.exec_info()[0]: {sys.exc_info()[0]}
        """
        msg.setText(output)
        msg.exec_()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config_file',
        help="please specify path to yaml config file", type=str,
        required=True,
        default=False)
    args = parser.parse_args()
    arguments = vars(args)
    with open(args.config_file, 'r') as f:
        config = yaml.safe_load(f)
    init_root_painter(config)



""" Utilities for working with the U-Net models 
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=C0111, R0913, R0914, W0511
import os
import time
import math
import traceback
import numpy as np
import torch
import copy
from skimage import img_as_float32
import im_utils
from unet3d import UNet3D, SmallUNet3D
from file_utils import ls
from torch.nn.functional import softmax
import torch.nn.functional as F
from loss import get_batch_loss

cached_model = None
cached_model_path = None
use_fake_cnn = False

def fake_cnn(tiles_for_gpu):
    """ Useful debug function for checking tile layout etc """
    output = []
    for t in tiles_for_gpu:
        v = t[0, 17:-17, 17:-17, 17:-17].data.cpu().numpy()
        v_mean = np.mean(v)
        output.append((v > v_mean).astype(np.int8))
    return np.array(output)
 
def get_in_w_out_w_pairs():
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # matching pairs of input/output sizes for specific unet used
    # 36 to 228 in incrememnts of 16 (sorted large to small)
    in_w_list = sorted([36 + (x*16) for x in range(20)], reverse=True)
    
    # output always 34 less than input
    out_w_list = [x - 34 for x in in_w_list]
    return list(zip(in_w_list, out_w_list))


def allocate_net(in_w, out_w, num_classes, small_unet):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    channels = 1 # change to 3 for auto-complete
    if small_unet:
        net = SmallUNet3D(im_channels=channels, num_classes=num_classes).to(device)
    else:
        net = UNet3D(im_channels=channels, num_classes=num_classes).to(device)
    net = torch.nn.DataParallel(net)
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01,
                                momentum=0.99, nesterov=True)
    net.train()
    for _ in range(3):
    
        #                      b, c,  d,  h,    w    
        input_data = np.zeros((4, channels, 52, in_w, in_w))
        optimizer.zero_grad()
        outputs = net(torch.from_numpy(input_data).to(device).float())
        batch_fg_tiles = torch.ones(4, num_classes, 52, in_w, in_w).long().to(device)
        batch_bg_tiles = torch.zeros(4, num_classes, 52, in_w, in_w).long().to(device)
        batch_fg_tiles[:, 0, 0] = 0
        batch_bg_tiles[:, 0, 0] = 1
        batch_classes = []
        for _ in range(4):
            batch_classes.append([f'c_{c}' for c in range(num_classes)])
        (batch_loss, _, _,
         _, _) = get_batch_loss(
             outputs, batch_fg_tiles, batch_bg_tiles,
             [[None for c in range(num_classes)] for t in batch_fg_tiles], # segmentation excluded from loss for now.
             batch_classes,
             [f'c_{c}' for c in range(num_classes)],
             compute_loss=True)
        batch_loss.backward()
        optimizer.step()
        del batch_loss
        del batch_fg_tiles
        del batch_bg_tiles
        del outputs
        del input_data


def get_in_w_out_w_for_memory(num_classes, small_unet):
    print('computing largest patch size for GPU, num class = ', num_classes)
    # search for appropriate input size for GPU
    # in_w, out_w = get_in_w_out_w_for_memory(num_classes)
    # try to train a network and see which patch size fits on the gpu.
    pairs = get_in_w_out_w_pairs()
    for i, (in_w, out_w) in enumerate(pairs):
        torch.cuda.empty_cache()
        try:
            allocate_net(in_w, out_w, num_classes, small_unet)
            torch.cuda.empty_cache()
            print(in_w, out_w, 'ok')
            print('using', pairs[i+1], 'to be safe') # return the next smallest to be safe
            return pairs[i+1] # return the next smallest to be safe
        except Exception as e:
            if 'out of memory' in str(e):
                print(in_w, out_w, 'too big')
            else:
                print(e, traceback.format_exc())
    raise Exception('Could not find patch small enough for available GPU memory')


def get_latest_model_paths(model_dir, k):
    fnames = ls(model_dir)
    fnames = sorted(fnames)[-k:]
    fpaths = [os.path.join(model_dir, f) for f in fnames]
    return fpaths


def load_model(model_path, classes, small_unet):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    global cached_model
    global cached_model_path
    
    # using cache can save up to half a second per segmentation with network drives
    if model_path == cached_model_path:
        return copy.deepcopy(cached_model)

    # two channels as one is input image and another is some of the fg and bg annotation
    # each non-empty channel in the annotation is included with 50% chance.
    # - fg and bg will go in as seprate channels 
    #  so channels are [image, fg_annot, bg_annot]
    if small_unet:
        model = SmallUNet3D(num_classes=len(classes), im_channels=1) # 3 channels as annotation may be used as input.
    else:
        model = UNet3D(num_classes=len(classes), im_channels=1)
    try:
        model.load_state_dict(torch.load(model_path, map_location=device))
        model = torch.nn.DataParallel(model)
    # pylint: disable=broad-except, bare-except
    except:
        model = torch.nn.DataParallel(model)
        model.load_state_dict(torch.load(model_path, map_location=device))
    if not use_fake_cnn:
        model.to(device)
    # store in cache as most frequest model is laoded often
    cached_model_path = model_path
    cached_model = model
    return copy.deepcopy(model)


def random_model(classes, small_unet):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # num out channels is twice number of channels
    # as we have a positive and negative output for each structure.
    # disabled for now as auto-complete feature is stalled.
    #model = UNet3D(classes, im_channels=3)
    if small_unet:
        model = SmallUNet3D(num_classes=len(classes), im_channels=1) # 3 channels as annotation may be used as input.
    else:
        model = UNet3D(num_classes=len(classes), im_channels=1)
    model = torch.nn.DataParallel(model)
    if not use_fake_cnn: 
        model.to(device)
    return model

def create_first_model_with_random_weights(model_dir, classes, small_unet):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # used when no model was specified on project creation.
    model_num = 1
    model_name = str(model_num).zfill(6)
    model_name += '_' + str(int(round(time.time()))) + '.pkl'
    model = random_model(classes, small_unet)
    model_path = os.path.join(model_dir, model_name)
    torch.save(model.state_dict(), model_path)
    if not use_fake_cnn: 
        model.to(device)
    return model


def get_prev_model(model_dir, classes, small_unet):
    prev_path = get_latest_model_paths(model_dir, k=1)[0]
    prev_model = load_model(prev_path, classes, small_unet)
    return prev_model, prev_path


def save_if_better(model_dir, cur_model, prev_model_path, cur_dice, prev_dice):
    # convert the nans as they don't work in comparison
    if math.isnan(cur_dice):
        cur_dice = 0
    if math.isnan(prev_dice):
        prev_dice = 0
    print('Validation: prev dice', str(round(prev_dice, 5)).ljust(7, '0'),
          'cur dice', str(round(cur_dice, 5)).ljust(7, '0'))
    if cur_dice > prev_dice:
        save_model(model_dir, cur_model, prev_model_path)
        return True
    return False

def save_model(model_dir, cur_model, prev_model_path):
    prev_model_fname = os.path.basename(prev_model_path)
    prev_model_num = int(prev_model_fname.split('_')[0])
    model_num = prev_model_num + 1
    now = int(round(time.time()))
    model_name = str(model_num).zfill(6) + '_' + str(now) + '.pkl'
    model_path = os.path.join(model_dir, model_name)
    print('saving', model_path, time.strftime('%H:%M:%S', time.localtime(now)))
    torch.save(cur_model.state_dict(), model_path)


def ensemble_segment_3d(model_paths, image, fname, batch_size, in_w, out_w, in_d,
                        out_d, classes, small_unet):
    """ Average predictions from each model specified in model_paths """
    t = time.time()
    input_image_shape = image.shape
    cnn = load_model(model_paths[0], classes, small_unet)
    in_patch_shape = (in_d, in_w, in_w)
    out_patch_shape = (out_d, out_w, out_w)

    depth_diff = in_patch_shape[0] - out_patch_shape[0]
    height_diff = in_patch_shape[1] - out_patch_shape[1]
    width_diff = in_patch_shape[2] - out_patch_shape[2]

    print('input image shape (before pad)= ', image.shape)
    # pad so seg will be size of input image
    image = im_utils.pad_3d(image, width_diff//2, depth_diff//2,
                            mode='reflect', constant_values=0)

    # segment returns a series of prediction maps. one for each class.
    print('input image shape (after pad)= ', image.shape)
    pred_maps = segment_3d(cnn, image, batch_size, in_patch_shape, out_patch_shape)

    print('pred maps[0].shape = ', pred_maps[0].shape)
    assert pred_maps[0].shape == input_image_shape, (
        f'pred_maps[0].shape: {pred_maps[0].shape}, '
        f'input_image_shape: {input_image_shape}')

    print('time to segment image', time.time() - t)
    return pred_maps


def segment_3d(cnn, image, batch_size, in_tile_shape, out_tile_shape, auto_complete_enabled=False):
    """
    in_tile_shape and out_tile_shape are (depth, height, width)
    """
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # Return prediction for each pixel in the image
    # The cnn will give a the output as channels where
    # each channel corresponds to a specific class 'probability'
    # don't need channel dimension
    # make sure the width, height and depth is at least as big as the tile.
    assert len(image.shape) == 3, str(image.shape)

    original_shape = image.shape

    # if the image is smaller than the patch size then pad it to be the same as the patch.
    padded_for_patch = False
    patch_pad_z = 0
    patch_pad_y = 0
    patch_pad_x = 0

    if image.shape[0] < in_tile_shape[0]:
        padded_for_patch = True
        patch_pad_z = in_tile_shape[0] - image.shape[0]

    if image.shape[1] < in_tile_shape[1]:
        padded_for_patch = True
        patch_pad_y = in_tile_shape[1] - image.shape[1]

    if image.shape[2] < in_tile_shape[2]:
        padded_for_patch = True
        patch_pad_x = in_tile_shape[2] - image.shape[2]

    if padded_for_patch:
        padded_image = np.zeros((
            image.shape[0] + patch_pad_z,
            image.shape[1] + patch_pad_y,
            image.shape[2] + patch_pad_x),
            dtype=image.dtype)
        padded_image[:image.shape[0], :image.shape[1], :image.shape[2]] = image
        image = padded_image  

    depth_diff = in_tile_shape[0] - out_tile_shape[0]
    width_diff = in_tile_shape[1] - out_tile_shape[1]
    
    out_im_shape = (image.shape[0] - depth_diff,
                    image.shape[1] - width_diff,
                    image.shape[2] - width_diff)

    coords = im_utils.get_coords_3d(out_im_shape, out_tile_shape)
    coord_idx = 0
    class_output_tiles = None # list of tiles for each class

    while coord_idx < len(coords):
        tiles_to_process = []
        coords_to_process = []
        for _ in range(batch_size):
            if coord_idx < len(coords):
                coord = coords[coord_idx]
                x_coord, y_coord, z_coord = coord
                tile = image[z_coord:z_coord+in_tile_shape[0],
                             y_coord:y_coord+in_tile_shape[1],
                             x_coord:x_coord+in_tile_shape[2]]

                # need to add channel dimension for GPU processing.
                tile = np.expand_dims(tile, axis=0)
                
                assert tile.shape[1] == in_tile_shape[0], str(tile.shape)
                assert tile.shape[2] == in_tile_shape[1], str(tile.shape)
                assert tile.shape[3] == in_tile_shape[2], str(tile.shape)

                tile = img_as_float32(tile)
                tile = im_utils.normalize_tile(tile)
                tile = img_as_float32(tile)
                coord_idx += 1
                tiles_to_process.append(tile) # need channel dimension
                coords_to_process.append(coord)

        tiles_to_process = np.array(tiles_to_process)
        tiles_for_gpu = torch.from_numpy(tiles_to_process)

        tiles_for_gpu = tiles_for_gpu.to(device).float()
        # TODO: consider use of detach. 
        # I might want to move to cpu later to speed up the next few operations.
        # I added .detach().cpu() to prevent a memory error.
        # pad with zeros for the annotation input channels
        # l,r, l,r, but from end to start     w  w  h  h  d  d, c, c, b, b
        if auto_complete_enabled:
            # add channels for annotation if auto_complete enabled
            tiles_for_gpu = F.pad(tiles_for_gpu, (0, 0, 0, 0, 0, 0, 0, 2), 'constant', 0)
        # tiles shape after padding torch.Size([4, 3, 52, 228, 228])
        outputs = cnn(tiles_for_gpu).detach().cpu()
        # bg channel index for each class in network output.
        class_idxs = [x * 2 for x in range(outputs.shape[1] // 2)]
        
        if class_output_tiles is None:
            class_output_tiles = [[] for _ in class_idxs]

        for i, class_idx in enumerate(class_idxs):
            class_output = outputs[:, class_idx:class_idx+2]
            # class_output : (batch_size, bg/fg, depth, height, width)
            softmaxed = softmax(class_output, 1) 
            foreground_probs = softmaxed[:, 1]  # just the foreground probability.
            predicted = foreground_probs > 0.5
            predicted = predicted.int()
            pred_np = predicted.data.cpu().numpy()
            for out_tile in pred_np:
                class_output_tiles[i].append(out_tile)

    class_pred_maps = []
    for i, output_tiles in enumerate(class_output_tiles):
        # reconstruct for each class
        reconstructed = im_utils.reconstruct_from_tiles(output_tiles,
                                                        coords, out_im_shape)
        if padded_for_patch:
            # go back to the original shape before padding.
            # what ever we added on to make it as big as the patch size
            # now take that away.
            reconstructed = reconstructed[:reconstructed.shape[0] - patch_pad_z,
                                          :reconstructed.shape[1] - patch_pad_y,
                                          :reconstructed.shape[2] - patch_pad_x]
        class_pred_maps.append(reconstructed)

    return class_pred_maps



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=C0111,E1102,C0103,W0703,W0511,E1136
import os
import glob
import shutil
import time
from functools import partial
from math import ceil
import random
import numpy as np
import skimage.util as skim_util
from skimage.exposure import rescale_intensity
import nibabel as nib
from file_utils import ls, ls_annotations
#import nrrd
from pathlib import Path
import traceback
import SimpleITK as sitk
from scipy.ndimage import zoom
import pandas as pd
import psycopg2


def get_recursive_files(input_path):
    filenames = []
    path = Path(input_path)
    for p in path.rglob("*"):
        if os.path.isdir(str(p)) is False:
            filenames.append(p._str)
    return filenames



def getDataFromDatabase(sql_config):
    connection = psycopg2.connect(
        host=sql_config['host'],
        database=sql_config['database'],
        user=sql_config['username'],
        password=sql_config['password'])
    sql = sql_config['query'].replace(
        "?table_name", "\"" + sql_config['table_name'] + "\"")
    sql = sql.replace(
        "?schema_name", "\"" + sql_config['schema_name'] + "\"")
    sql = sql.replace(
        "??", "\"")
    df = pd.read_sql_query(sql, connection)
    if len(df) == 0:
        print('The requested query does not have any data!')
    connection.close()

    return df['DcmPathFlatten'].tolist()

def is_image(fname):
    """ extensions that have been tested with so far """
    extensions = {".jpg", ".png", ".jpeg", '.tif', '.tiff'}
    fname_ext = os.path.splitext(fname)[1].lower()
    return (fname_ext in extensions or fname.endswith('.nii.gz') or 
            fname.endswith('.npy') or fname.endswith('.nrrd') or fname.endswith('.dcm')
            or fname.endswith('.DCM'))

def normalize_tile(tile):
    if np.min(tile) < np.max(tile):
        tile = rescale_intensity(tile, out_range=(0, 1))
    else:
        # a single value tile is very rare but it has been claimed
        # that it may occur in X-ray background regions.
        # set to be 0 to ensure it is within a similar range to the normalized tiles.
        tile *= 0 
    assert np.min(tile) >= 0, f"tile min {np.min(tile)}"
    assert np.max(tile) <= 1, f"tile max {np.max(tile)}"
    return tile


def reconstruct_from_tiles(tiles, coords, output_shape):
    image = np.zeros(output_shape)
    for tile, (x_coord, y_coord, z_coord) in zip(tiles, coords):
        image[z_coord:z_coord+tile.shape[0],
              y_coord:y_coord+tile.shape[1],
              x_coord:x_coord+tile.shape[2]] = tile
    return image


def load_with_retry(load_fn, fpath):
    max_attempts = 2
    attempts = 0
    while attempts < max_attempts:
        attempts += 1
        # file systems are unpredictable.
        # We may have problems reading the file.
        # try-catch to avoid this.
        # (just try again)
        try:
            image = load_fn(fpath)
            return image
        except Exception as e:
            print('load_with_retry', fpath, 'exception', e, traceback.format_exc())
            # This could be due to an empty annotation saved by the user.
            # Which happens rarely due to deleting all labels in an
            # existing annotation and is not a problem.
            # give it some time and try again.
            time.sleep(0.1)
    if attempts == max_attempts:
        raise Exception('Could not load. Too many retries')



def load_image_and_annot_for_seg(dataset_dir, train_annot_dirs, fname):
    """
    Load image and annotation to enable segmentation.

    returns
        image (np.array) - image data
        annots (list(np.array)) - annotations associated with fname
        classes (list(string)) - classes for each annot,
                                 taken from annot directory name
        fname - file name
    """
    def load_fname(train_annot_dirs, dataset_dir, fname):
        # This might take ages, profile and optimize
        fnames = []
        # each annotation corresponds to an individual class.
        all_classes = []
        all_dirs = []
        
       
        for train_annot_dir in train_annot_dirs:
            annot_fnames = ls_annotations(train_annot_dir)
            fnames += annot_fnames
            # Assuming class name is in annotation path
            # i.e annotations/{class_name}/train/annot1.png,annot2.png..
            class_name = Path(train_annot_dir).parts[-2]
            all_classes += [class_name] * len(annot_fnames)
            all_dirs += [train_annot_dir] * len(annot_fnames)

        # triggers retry if assertion fails
        assert is_image(fname), f'{fname} is not a valid image'

        # annots and classes associated with fname
        indices = [i for i, f in enumerate(fnames) if f == fname]
        classes = [all_classes[i] for i in indices]
        annot_dirs = [all_dirs[i] for i in indices]
        annots = []

        for annot_dir in annot_dirs:
            annot_path = os.path.join(annot_dir, fname)
            annot = load_image(annot_path)
            # Why would we have annotations without content?
            assert np.sum(annot) > 0
            annot = np.pad(annot, ((0, 0), (17,17), (17,17), (17, 17)), mode='constant')
            annots.append(annot)

        # it's possible the image has a different extenstion
        # so use glob to get it
        fname_no_ext = fname.replace('.nii.gz', '').replace('.nrrd', '').replace('.dcm', '')
        image_path_part = os.path.join(dataset_dir, fname_no_ext)
        image_path = glob.glob(image_path_part + '.*')[0]
        image = load_image(image_path)

        #  needs to be swapped to channels first and rotated etc
        # to be consistent with everything else.
        # todo: consider removing this soon.
        #image = np.rot90(image, k=3)
        #image = np.moveaxis(image, -1, 0) # depth moved to beginning
        # reverse lr and ud
       # image = image[::-1, :, ::-1]

        # also return fname for debugging purposes.
        return image, annots, classes, fname

    load_fname = partial(load_fname, train_annot_dirs, dataset_dir)
    return load_with_retry(load_fname, fname)



def load_train_image_and_annot(dataset_dir, train_seg_dirs, train_annot_dirs):
    """
    returns
        image (np.array) - image data
        annots (list(np.array)) - annotations associated with fname
        segs (list(np.array)) - segmentations associated with fname
        classes (list(string)) - classes for each annot,
                                 taken from annot directory name
        fname - file name
    """

    def load_random(train_annot_dirs, train_seg_dirs, dataset_dir, _):
        # This might take ages, profile and optimize
        fnames = []
        # each annotation corresponds to an individual class.
        all_classes = []
        all_annot_dirs = []
        all_seg_dirs = []

        train_seg_dirs = sorted(train_seg_dirs)
        train_annot_dirs = sorted(train_annot_dirs)
        assert len(train_seg_dirs) == len(train_annot_dirs)
    
        for train_seg_dir, train_annot_dir in zip(train_seg_dirs, train_annot_dirs):
            annot_fnames = ls_annotations(train_annot_dir)
            fnames += annot_fnames
            # Assuming class name is in annotation path
            # i.e annotations/{class_name}/train/annot1.png,annot2.png..
            class_name = Path(train_annot_dir).parts[-2]
            all_classes += [class_name] * len(annot_fnames)
            all_annot_dirs += [train_annot_dir] * len(annot_fnames)
            all_seg_dirs += [train_seg_dir] * len(annot_fnames)
        
        assert len(fnames), 'should be at least one fname'

        fname = random.sample(fnames, 1)[0]

        # triggers retry if assertion fails
        assert is_image(fname), f'{fname} is not a valid image'

        # annots and classes associated with fname
        indices = [i for i, f in enumerate(fnames) if f == fname]
        classes = [all_classes[i] for i in indices]
        annot_dirs = [all_annot_dirs[i] for i in indices]
        seg_dirs = [all_seg_dirs[i] for i in indices]
        annots = []
        segs = []

        for annot_dir, seg_dir in zip(annot_dirs, seg_dirs): # for each of the classes.
            annot_path = os.path.join(annot_dir, fname)
            annot = load_image(annot_path)
            # TODO not pad here?
            annot = pad_annot(annot, 34)

         #   annot = pad_annot(annot, 34) # removed 22nov
            # Why would we have annotations without content?
            assert np.sum(annot) > 0
            annot = np.pad(annot, ((0, 0), (17,17), (17,17), (17, 17)), mode='constant')
            annots.append(annot)

            seg_path = os.path.join(seg_dir, fname)
            if os.path.isfile(seg_path):
                seg = load_image(seg_path)
                # TODO :
                # MAYBE NOT (0, 0) ???
              #  seg = pad_image(seg, 34) # removed 22nov
                seg = pad_image(seg, 34)
                seg = np.pad(seg, ((17,17), (17,17), (17, 17)), mode='constant')
            else:
                seg = None
            segs.append(seg)

        # it's possible the image has a different extenstion
        # so use glob to get it
        fname_no_ext = fname.replace('.nii.gz', '').replace('.nrrd', '').replace('.dcm', '')
        image_path_part = os.path.join(dataset_dir, fname_no_ext)
        image_path = glob.glob(image_path_part + '.*')[0]
        image = load_image(image_path)
      #  image = pad_image(image, 34) # removed 22nov
        #  needs to be swapped to channels first and rotated etc
        # to be consistent with everything else.
        # todo: consider removing this soon.
        #image = np.rot90(image, k=3)
       # image = np.moveaxis(image, -1, 0) # depth moved to beginning
        # reverse lr and ud
      #  image = image[::-1, :, ::-1]
        # images are no longer padded on disk
        image = np.pad(image, ((17,17), (17,17), (17, 17)), mode='constant')
        # also return fname for debugging purposes.
        return image, annots, segs, classes, fname

    load_random = partial(load_random, train_annot_dirs, train_seg_dirs, dataset_dir)
    return load_with_retry(load_random, None)

def pad_3d(image, width, depth, mode='reflect', constant_values=0):
    pad_shape = [(depth, depth), (width, width), (width, width)]
    if len(image.shape) == 4:
        # assume channels first for 4 dimensional data.
        # don't pad channels
        pad_shape = [(0, 0)] + pad_shape
    if mode == 'reflect':
        return np.pad(image, pad_shape, mode)
    return np.pad(image, pad_shape, mode=mode,
                  constant_values=constant_values)


def get_val_tile_refs(annot_dirs, prev_tile_refs, out_shape):
    """
    Get tile info which covers all annotated regions of the annotation dataset.
    The list must be structured such that an index can be used to refer to each example
    so that it can be used with a dataloader.

    returns tile_refs (list)
        Each element of tile_refs is a list that includes:
            * image file name (string) - for loading the image from disk during validation
            * coord (x int, y int) - for addressing the location within the padded image
            * annot_mtime (int)
                The image annotation may get updated by the user at any time.
                We can use the mtime to check for this.
                If the annotation has changed then we need to retrieve tile
                coords for this image again. The reason for this is that we
                only want tile coords with annotations in. The user may have added or removed
                annotation in part of an image. This could mean a different set of coords (or
                not) should be returned for this image.

    Parameter prev_tile_refs is used for comparing both file names and mtime.

    The annot_dir folder should be checked for any new files (not in
    prev_tile_refs) or files with an mtime different from prev_tile_refs. For these
    file, the image should be loaded and new tile_refs should be retrieved. For all
    other images the tile_refs from prev_tile_refs can be used.
    """
    tile_refs = []

    # TODO, change this so we have a list of all annot names and their corresopnding classes.
    #  make a large list of classes that corresponds to all the file names. Ive done this elsewhere.

    # This might take ages, profile and optimize
    cur_annot_fnames = []
    # each annotation corresponds to an individual class.
    all_classes = []
    all_dirs = []

    for annot_dir in annot_dirs:
        annot_fnames = ls(annot_dir)
        if annot_fnames:
            cur_annot_fnames += annot_fnames
            # Assuming class name is in annotation path
            # i.e annotations/{class_name}/train/annot1.png,annot2.png..
            class_name = Path(annot_dir).parts[-2]
            all_classes += [class_name] * len(annot_fnames)
            all_dirs += [annot_dir] * len(annot_fnames)
    
    prev_annot_fnames = [r[0] for r in prev_tile_refs]
    all_annot_fnames = set(cur_annot_fnames + prev_annot_fnames)

    for annot_dir, annot_fname in zip(all_dirs, all_annot_fnames):
        # get existing coord refs for this image
        prev_refs = [r for r in prev_tile_refs if r[0] == annot_fname]
        prev_mtimes = [r[2] for r in prev_tile_refs if r[0] == annot_fname]
        need_new_refs = False
        # if no refs for this image then check again
        if not prev_refs:
            need_new_refs = True
        else:
            annot_path = os.path.join(annot_dir, annot_fname)
            # if the file no longer exists then we do need new refs
            # surprisingly this did happen, I presume the file list was somehow out of date
            # and the removal of the file was only detected when trying to read it.
            if not os.path.isfile(annot_path):
                need_new_refs = True
            else:
                # otherwise check the modified time of the refs against the file.
                prev_mtime = prev_mtimes[0]
                cur_mtime = os.path.getmtime(os.path.join(annot_dir, annot_fname))

                # if file has been updated then get new refs
                if cur_mtime > prev_mtime:
                    need_new_refs = True
        if need_new_refs:
            new_file_refs = get_val_tile_refs_for_annot_3d(annot_dir, annot_fname, out_shape)
            tile_refs += new_file_refs
        else:
            tile_refs += prev_refs
    return tile_refs


def get_val_tile_refs_for_annot_3d(annot_dir, annot_fname, out_shape):
    """
    Each element of tile_refs is a list that includes:
        * image file name (string) - for loading the image from disk during validation
        * coord (x int, y int) - for addressing the location within the padded image
        * annot_mtime (int)
        * cached performance for this tile with previous (current best) model.
          Initialized to None but otherwise [tp, fp, tn, fn]
    """
    annot_path = os.path.join(annot_dir, annot_fname)
    if not os.path.isfile(annot_path):
        return []
    annot = load_image(annot_path)
    new_file_refs = []
    annot_shape = annot.shape[1:]
    coords = get_coords_3d(annot_shape, out_tile_shape=out_shape)

    mtime = os.path.getmtime(annot_path)
    for (x, y, z) in coords:
        annot_tile = annot[:, z:z+out_shape[0], y:y+out_shape[1], x:x+out_shape[2]]
        # we only want to validate on annotation tiles
        # which have annotation information.
        if np.any(annot_tile):
            # fname, [x, y, z], mtime, prev model metrics i.e [tp, tn, fp, fn] or None
            new_file_refs.append([annot_fname, [x, y, z], mtime, None])
    return new_file_refs


def get_coords_3d(annot_shape, out_tile_shape):
    """ Get the coordinates relative to the output image for the 
        validation routine. These coordinates will lead to patches
        which cover the image with minimum overlap (assuming fixed size patch) """

    assert len(annot_shape) == 3, str(annot_shape) # d, h, w
    
    depth_count = ceil(annot_shape[0] / out_tile_shape[0])
    vertical_count = ceil(annot_shape[1] / out_tile_shape[1])
    horizontal_count = ceil(annot_shape[2] / out_tile_shape[2])

    # first split the image based on the tiles that fit
    z_coords = [d*out_tile_shape[0] for d in range(depth_count-1)] # z is depth
    y_coords = [v*out_tile_shape[1] for v in range(vertical_count-1)]
    x_coords = [h*out_tile_shape[2] for h in range(horizontal_count-1)]

    # The last row and column of tiles might not fit
    # (Might go outside the image)
    # so get the tile positiion by subtracting tile size from the
    # edge of the image.
    lower_z = annot_shape[0] - out_tile_shape[0]
    bottom_y = annot_shape[1] - out_tile_shape[1]
    right_x = annot_shape[2] - out_tile_shape[2]

    z_coords.append(max(0, lower_z))
    y_coords.append(max(0, bottom_y))
    x_coords.append(max(0, right_x))

    # because its a cuboid get all combinations of x, y and z
    tile_coords = [(x, y, z) for x in x_coords for y in y_coords for z in z_coords]
    return tile_coords



def save_then_move(out_path, seg):
    """ need to save in a temp folder first and
        then move to the segmentation folder after saving
        this is because scripts are monitoring the segmentation folder
        and the file saving takes time..
        We don't want the scripts that monitor the segmentation
        folder to try loading the file half way through saving
        as this causes errors. Thus we save and then rename.
    """
    raise Exception('Depracated')
    fname = os.path.basename(out_path)
    token = str(time.time()) # add token to avoid resaving over files wiith the same name
    temp_path = os.path.join('/tmp', token + fname)
    if out_path.endswith('.nii.gz'):
        img = nib.Nifti1Image(seg, np.eye(4))
        img.to_filename(temp_path)
    elif out_path.endswith('.npy'):
        np.save(temp_path, seg)
    else:
        raise Exception(f'Unhandled {out_path}')
    shutil.copy(temp_path, out_path)
    os.remove(temp_path)


def save(out_path, seg):
    dir_to_create = os.path.dirname(out_path)
    Path(dir_to_create).mkdir(parents=True, exist_ok=True)

    if out_path.endswith('.nii.gz'):
        print('save seg shape', seg.shape)
        seg = seg[::-1, :, ::-1]
        img = nib.Nifti1Image(seg, np.eye(4))
        
        img.to_filename(out_path)
    else:
        raise Exception(f'Unhandled {out_path}')

def resizeVolume(img, output_size):
    factors = (output_size[0]/img.shape[1],
                output_size[1]/img.shape[2])

    return zoom(img, (1, factors[0], factors[1]))

def load_image(image_path):
    if image_path.endswith('.npy'):
        image = np.load(image_path, mmap_mode='c')
    elif image_path.endswith('.nrrd'):
        raise ValueError('not implemented')
        #image, _ = nrrd.read(image_path)
    elif image_path.endswith('.nii.gz'):
        # We don't currently use them during training but it's useful to be
        # able to load nifty files directory to give the user
        # more convenient segmentation options.
        image = nib.load(image_path)
        image = np.array(image.dataobj)
       # print('ups this function was not supoosed to run')
    elif image_path.endswith(('.dcm', '.DCM')):
        image = sitk.ReadImage(image_path)
        image = sitk.GetArrayFromImage(image)
   #     image = resizeVolume(image, (128, 128))
      #  image = pad_image(image, 34)
        
    else:
        print('image path', image_path)
        raise ValueError('not implemented')
    return image

def pad_image(image, pad_size):
    shap = image.shape
    if shap[0] < pad_size:
        to_pad = pad_size - shap[0]
        if (to_pad % 2) == 0:
            image = np.pad(image, ((int(to_pad/2),int(to_pad/2)), (0,0), (0, 0)), 'constant')
        else:
            image = np.pad(image, ((int(to_pad/2)+1, int(to_pad/2)), (0,0), (0, 0)), 'constant')
    # else:
    #     image = image[0:pad_size, :, :]
    return image

def pad_annot(image, pad_size):
    shap = image.shape
    if shap[1] < pad_size:
        to_pad = pad_size - shap[1]
        if (to_pad % 2) == 0:
            image = np.pad(image, ((0,0), (int(to_pad/2),int(to_pad/2)), (0,0), (0, 0)), 'constant') # 'reflect'
        else:
            image = np.pad(image, ((0,0), (int(to_pad/2)+1, int(to_pad/2)), (0,0), (0, 0)), 'constant') # 'reflect'
    # else:
    #     image = image[0:pad_size, :, :]
    return image


from .trainer import Trainer
from .loss import *
from .data_utils import *
from .datasets import *
from .file_utils import *
from .instructions import *
from .metrics import *
from .model_utils import *
from .patch_seg import *
from .scp_utils import *
from .startup import *
from .unet3d import *
import importlib
import pkgutil

# Get the current package name
package_name = __name__

# Iterate through all the modules in the package
for _, module_name, _ in pkgutil.iter_modules(__path__):
    # Import the module
    module = importlib.import_module(f"{package_name}.{module_name}")

    # Iterate through all the attributes in the module
    for attr_name in dir(module):
        # Get the attribute
        attr = getattr(module, attr_name)

        # Check if the attribute is a class or a function
        if callable(attr) and not attr_name.startswith("__"):
            # Import the attribute into the package's namespace
            globals()[attr_name] = attr



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui
import numpy as np


class ContrastSlider(QtWidgets.QWidget):
    changed = QtCore.pyqtSignal(int, int)

    def __init__(self, presets):
        super().__init__()
        # min HU, max HU, brightness %
        self.presets = presets
        preset = self.presets[list(self.presets.keys())[0]] 
        self.hu_range = [preset[0], preset[1]]
        self.min_value = preset[0]
        self.max_value = preset[1]
        self.brightness_value = round(preset[2])
        self.initUI()

    def update_range(self, img):
        min_hu = np.min(img)
        max_hu = np.max(img)
        self.hu_range = [min_hu, max_hu]
        self.min_slider.setMinimum(self.hu_range[0])
        self.min_slider.setMaximum(self.hu_range[1])
        self.max_slider.setMinimum(self.hu_range[0])
        self.max_slider.setMaximum(self.hu_range[1])
        self.min_value_label.setText(str(self.min_value) + ' HU')
        self.max_value_label.setText(str(self.max_value) + ' HU')

    def initUI(self):
        self.layout = QtWidgets.QVBoxLayout()
        self.debounce = QtCore.QTimer()
        self.debounce.setInterval(5)
        self.debounce.setSingleShot(True)
        self.debounce.timeout.connect(self.debounced)

        self.min_slider_container = QtWidgets.QWidget()
        self.min_slider_layout = QtWidgets.QHBoxLayout()
        self.min_slider_layout.addWidget(QtWidgets.QLabel("Min:"))
        self.min_slider_container.setLayout(self.min_slider_layout)
        self.min_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.min_slider.setMinimum(self.hu_range[0])
        self.min_slider.setMaximum(self.hu_range[1])
        self.min_slider.setValue(self.min_value)
        self.min_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.min_slider.setTickInterval((round((self.hu_range[1] - self.hu_range[0]) / 100)))
        self.min_slider.setFixedWidth(300)
        self.min_slider.valueChanged.connect(self.value_changed)
        self.min_slider_layout.addWidget(self.min_slider)
        self.min_value_label = QtWidgets.QLabel(str(self.min_value) + ' HU')
        self.min_slider_layout.addWidget(self.min_value_label)
        self.layout.addWidget(self.min_slider_container)

        self.max_slider_container = QtWidgets.QWidget()
        self.max_slider_layout = QtWidgets.QHBoxLayout()
        self.max_slider_layout.addWidget(QtWidgets.QLabel("Max:"))
        self.max_slider_container.setLayout(self.max_slider_layout)
        self.max_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.max_slider.setMinimum(self.hu_range[0])
        self.max_slider.setMaximum(self.hu_range[1])
        self.max_slider.setValue(self.max_value)
        self.max_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.min_slider.setTickInterval((round((self.hu_range[1] - self.hu_range[0]) / 100)))
        self.max_slider.setFixedWidth(300)
        self.max_slider.valueChanged.connect(self.value_changed)
        self.max_slider_layout.addWidget(self.max_slider)
        self.max_value_label = QtWidgets.QLabel(str(self.max_value) + ' HU')
        self.max_slider_layout.addWidget(self.max_value_label)
        self.layout.addWidget(self.max_slider_container)


        self.brightness_slider_container = QtWidgets.QWidget()
        self.brightness_slider_layout = QtWidgets.QHBoxLayout()
        self.brightness_slider_layout.addWidget(QtWidgets.QLabel("Brightness:"))
        self.brightness_slider_container.setLayout(self.brightness_slider_layout)
        self.brightness_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.brightness_slider.setMinimum(0)
        self.brightness_slider.setMaximum(300)
        self.brightness_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.brightness_slider.setTickInterval(1)
        self.brightness_slider.setFixedWidth(300)
        self.brightness_slider.setValue(self.brightness_value)
        self.brightness_slider.valueChanged.connect(self.value_changed)
        self.brightness_slider_layout.addWidget(self.brightness_slider)
        self.brightness_value_label = QtWidgets.QLabel(str(self.brightness_value) + ' %')
        self.brightness_slider_layout.addWidget(self.brightness_value_label)
        self.layout.addWidget(self.brightness_slider_container)

        self.setLayout(self.layout)
        self.setWindowTitle("Contrast Settings")
        self.setFixedWidth(525)

        self.preset_container = QtWidgets.QWidget()
        self.preset_layout = QtWidgets.QHBoxLayout()
        self.preset_container.setLayout(self.preset_layout)
        self.preset_layout.addWidget(QtWidgets.QLabel("Select preset:"))

        presets_combo = QtWidgets.QComboBox(self)
        self.presets_combo = presets_combo
        for preset_name in self.presets:
            presets_combo.addItem(preset_name)
        presets_combo.activated[str].connect(self.preset_selected)
        self.preset_layout.addWidget(presets_combo)
        self.layout.addWidget(self.preset_container)

    def preset_selected(self, preset_name):
        preset = self.presets[preset_name]
        self.min_slider.setValue(preset[0])
        self.max_slider.setValue(preset[1])
        self.brightness_slider.setValue(preset[2])
        self.presets_combo.setCurrentText(preset_name)
        self.value_changed()

    def value_changed(self):
        self.min_value = self.min_slider.value()
        self.min_value_label.setText(str(self.min_value) + ' HU')
        self.max_value = self.max_slider.value()
        self.max_value_label.setText(str(self.max_value) + ' HU')
        self.brightness_value = self.brightness_slider.value()
        self.brightness_value_label.setText(str(self.brightness_value) + ' %')
        self.debounce.start()

    def debounced(self):
        self.changed.emit(self.min_value, self.max_value)



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import shutil
from os.path import join

def fix_app():
    """
    Unfortunately fbs/pyInstaller is no longer capable of building
    RootPainter on it's own.

    It may be that I upgraded library versions to ones which are no longer
    supported i.e scikit-image.
    
    We will need to copy accross some dependencies to get it working again.
    """
    
    # If you are using Mac, then you will have the following folder after
    # running the fbs freeze command
    # ./dist/RootPainter.app
    is_mac = os.path.isdir('./dist/RootPainter3D.app')
    is_linux = os.path.exists('./dist/RootPainter3D/RootPainter3D')
    
    # or the following folder on windows
    is_windows = os.path.exists('dist\RootPainter3D\RootPainter3D.exe')
   
    print('is_windows', is_windows)
    print('is_mac', is_mac)
    print('is_linux', is_linux)

    # If you try to run RootPainter on the command line like so:
    # ./dist/RootPainter.app/Contents/MacOS/RootPainter 
    # Then you may receive the following error:
    # File "skimage/feature/orb_cy.pyx", line 12, in init skimage.feature.orb_cy
    # ModuleNotFoundError: No module named 'skimage.feature._orb_descriptor_positions'
    
    # It seems the built application is missing some crucial files from skimage.
    # To copy these accross we will assume you have an environment created with venv (virtual env)
    # in the current working directory call 'env'
    env_dir = './env'
    assert os.path.isdir(env_dir), f'Could not find env folder {env_dir}'
    if is_mac:
        site_packages_dir = os.path.join(env_dir, 'lib/python3.9/site-packages')
        build_dir = './dist/RootPainter3D.app/Contents/MacOS/'
    elif is_windows:
        site_packages_dir = os.path.join(env_dir, 'Lib', 'site-packages')
        build_dir = './dist/RootPainter3D'
    elif is_linux:
        site_packages_dir = os.path.join(env_dir, 'lib/python3.10/site-packages')
        build_dir = './dist/RootPainter3D'


    # Copy missing orb files
    skimage_dir = join(site_packages_dir, 'skimage')

    orbpy_src = join(skimage_dir, 'feature/_orb_descriptor_positions.py')
    orbpy_target = join(build_dir, 'skimage/feature/_orb_descriptor_positions.py')
    shutil.copyfile(orbpy_src, orbpy_target)

    # copy missing orb plugin file
    orbtxt_src = join(skimage_dir, 'feature/orb_descriptor_positions.txt')
    orbtxt_target = join(build_dir, 'skimage/io/_plugins/orb_descriptor_positions.txt')
    shutil.copyfile(orbtxt_src, orbtxt_target)


    # Copy missing tiffile plugin
    tif_src = join(skimage_dir, 'io/_plugins/tifffile_plugin.py')
    tif_target = join(build_dir, 'skimage/io/_plugins/tifffile_plugin.py')
    shutil.copyfile(tif_src, tif_target)

    for dir_name in ['filters', 'metrics', 'segmentation', 'restoration']:
        if not os.path.isdir(os.path.join(build_dir, 'skimage', dir_name)):
            os.makedirs(os.path.join(build_dir, 'skimage', dir_name))
        for f in os.listdir(os.path.join(skimage_dir, dir_name)):
            src = join(skimage_dir, dir_name, f)
            target = join(build_dir, 'skimage', dir_name, f)
            if not os.path.isfile(target) and not os.path.isdir(src):
                shutil.copyfile(src, target)

if __name__ == '__main__':
    fix_app()




"""
Copyright (C) 2019 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
# torch has no sum member
# pylint: disable=E1101

import torch
from torch.nn.functional import softmax
from torch.nn.functional import cross_entropy
from torch.nn.functional import l1_loss
import numpy as np


def dice_loss(predictions, labels):
    """ based on loss function from V-Net paper """
    softmaxed = softmax(predictions, 1)
    predictions = softmaxed[:, 1, :]  # just the root probability.
    labels = labels.float()
    preds = predictions.contiguous().view(-1)
    labels = labels.view(-1)
    intersection = torch.sum(torch.mul(preds, labels))
    union = torch.sum(preds) + torch.sum(labels)
    return 1 - ((2 * intersection) / (union))


def combined_loss(predictions, labels):
    """ mix of dice and cross entropy """
    # if they are bigger than 1 you get a strange gpu error
    # without a stack track so you will have no idea why.
    assert torch.max(labels) <= 1
    if torch.sum(labels) > 0:
        return (dice_loss(predictions, labels) +
                (0.3 * cross_entropy(predictions, labels)))
    # When no roots use only cross entropy
    # as dice is undefined.
    return 0.3 * cross_entropy(predictions, labels)


def get_batch_loss(outputs, batch_fg_tiles, batch_bg_tiles, batch_seg_tiles,
                   batch_classes, project_classes,
                   compute_loss):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    """

        outputs - predictions from neural network (not softmaxed)
        batch_fg_tiles - list of tiles, each tile is binary map of foreground annotation
        batch_bg_tiles - list of tiles, each tile is binary map of background annotation
        compute_loss - can be false if only metrics are required.

        returns
            batch_loss - loss used to update the network
            tps - true positives for batch
            tns - true negatives for batch
            fps - false positives for batch
            fns - false negatives for batch
            defined_total - number of pixels with annotation defined.
    """

    defined_total = 0
    class_losses = [] # loss for each class

    instance_tps = [0 for _ in range(outputs.shape[0])]
    instance_tns = list(instance_tps)
    instance_fps = list(instance_tps)
    instance_fns = list(instance_tps)

    for unique_class in project_classes:

        # for each class we need to get a tensor with shape
        # [batch_size, 2 (bg,fg), h, w]

        # fg,bg pairs related to this class for each im_tile in the batch
        class_outputs = []

        # The fg tiles (ground truth)
        fg_tiles = []
        masks = [] # and regions of the image that were annotated.
        seg_tiles = []
        seg_class_outputs = []
        
        # go through each instance in the batch.
        for im_idx in range(outputs.shape[0]):

            # go through each class for this instance.
            for i, classname in enumerate(batch_classes[im_idx]):
                
                # if the classname is the class we are interested in
                if classname == unique_class:

                    # foregorund and background channels
                    fg_tile = batch_fg_tiles[im_idx][i][17:-17,17:-17,17:-17]
                    bg_tile = batch_bg_tiles[im_idx][i][17:-17,17:-17,17:-17]
                    mask = fg_tile + bg_tile
                    class_idx = project_classes.index(classname) * 2 # posiion in output.
                    class_output = outputs[im_idx][class_idx:class_idx+2]
    


                    mask = mask.to(device)
                    fg_tile = fg_tile.to(device)

                    # I want to get tps, tns, fps and fns 
                    # from fg_tile, mask, class_outputs
                    softmaxed = softmax(class_output, 0)
                    fg_prob = softmaxed[1]


                    assert fg_prob.shape == mask.shape, (
                        f"fg_prob shape {fg_prob.shape} and mask shape {mask.shape}"
                        f"should be equal")
                    fg_prob = fg_prob * mask
                    class_pred = fg_prob > 0.5   
                    class_pred = class_pred[mask > 0]
                    fg = fg_tile[mask > 0]
                    
                    instance_tps[im_idx] += torch.sum((fg == 1) * (class_pred == 1)).cpu().numpy()
                    instance_tns[im_idx] += torch.sum((fg == 0) * (class_pred == 0)).cpu().numpy()
                    instance_fps[im_idx] += torch.sum((fg == 0) * (class_pred == 1)).cpu().numpy()
                    instance_fns[im_idx] += torch.sum((fg == 1) * (class_pred == 0)).cpu().numpy()

                    masks.append(mask)
                    fg_tiles.append(fg_tile)
                    class_outputs.append(class_output)

                    seg_tile = batch_seg_tiles[im_idx][i]
                    if seg_tile is not None:
                        seg_tile = torch.from_numpy(seg_tile[17:-17,17:-17,17:-17]).to(device)
                        # for this purpose we ensure segmentation never disagrees with annotation.
                        seg_tile[fg_tile] = 1
                        seg_tile[bg_tile] = 0
                        seg_tiles.append(seg_tile)
                        seg_class_outputs.append(class_output)
             

        if not len(fg_tiles):
            continue
        
        if compute_loss:
            fg_tiles = torch.stack(fg_tiles).to(device)
            masks = torch.stack(masks).to(device)
            class_outputs = torch.stack(class_outputs)
                        
            if len(seg_class_outputs):

                seg_class_outputs = torch.stack(seg_class_outputs)
                seg_class_outputs_softmaxed = softmax(seg_class_outputs, 1)
                seg_foreground_probs = seg_class_outputs_softmaxed[:, 1]
                seg_tiles = torch.stack(seg_tiles).long()
                # See loss functions in https://arxiv.org/pdf/1912.02911.pdfhttps://arxiv.org/pdf/1912.02911.pdf
                class_seg_loss = l1_loss(seg_foreground_probs, seg_tiles, reduction='mean') # MAE
                class_losses.append(class_seg_loss)

            # remove any of the predictions for which we don't have ground truth
            # Set outputs to 0 where annotation undefined so that
            # The network can predict whatever it wants without any penalty.
            class_outputs[:, 0] *= masks
            class_outputs[:, 1] *= masks
            class_loss = combined_loss(class_outputs, fg_tiles)
            class_losses.append(class_loss)

    if compute_loss:
        return torch.mean(torch.stack(class_losses)), instance_tps, instance_tns, instance_fps, instance_fns
    return None, instance_tps, instance_tns, instance_fps, instance_fns



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=I1101, C0111, E0611, R0902
""" Canvas where image and annotations can be drawn """
import os
import numpy as np
from PyQt5 import QtWidgets
from PyQt5 import QtGui
from PyQt5 import QtCore
from PyQt5.QtCore import Qt
import nibabel as nib
import time
import qimage2ndarray
from skimage.segmentation import flood
import im_utils
from view_state import ViewState
from patch_seg import SegmentPatchThread, PatchSegmentor

class GraphicsScene(QtWidgets.QGraphicsScene):
    """
    Canvas where image and lines will be drawn
    """
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        if self.parent.parent.auto_complete_enabled:
            self.patch_segmentor = PatchSegmentor(self.parent.parent)
        self.regions_to_restrict_to = []
        self.cursor_shown = False
        self.brush_size = 25
        # history is a list of pixmaps
        self.history = []
        self.redo_list = []
        self.last_x = None
        self.last_y = None
        self.annot_pixmap = None
        self.outline_pixmap = None
        self.cursor_pixmap_holder = None
        self.outline_pixmap_holder = None
        self.cursor_pixmap = None
        self.mouse_down = False
 
    def keyReleaseEvent(self, event):
        if event.key() == QtCore.Qt.Key_Alt:
            # then remove everything except for the clicked regions
            n = len(self.regions_to_restrict_to)
            # dont do anything unless some regions were selected.
            if n > 0:
                button_reply = QtWidgets.QMessageBox.question(self.parent,
                    'Confirm',
                    "Are you sure you want to assign corrections to restrict" 
                    f" to the {n} clicked 3D object(s)? This action cannot be undone.",
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Cancel, 
                    QtWidgets.QMessageBox.Cancel)

                if button_reply == QtWidgets.QMessageBox.Yes:
                    self.parent.parent.info_label.setText("Removing disconnected regions")
                    idx = self.parent.slice_nav.max_slice_idx - self.parent.cur_slice_idx 
                    # correct to restrict to only the selected region
                    new_annot, removed_count, holes, error = im_utils.restrict_to_regions_containing_points(
                        self.parent.parent.seg_data,
                        self.parent.parent.annot_data,
                        self.regions_to_restrict_to)
                    if error:
                        self.parent.parent.info_label.setText(error)
                    else:
                        message = f"{removed_count} foreground regions and {holes} holes were removed"
                        self.parent.parent.info_label.setText(message)
                    self.parent.parent.annot_data = new_annot
                    self.parent.parent.update_viewer_annot_slice()
                    self.parent.parent.update_viewer_outline()

                    self.regions_to_restrict_to = []
        super().keyReleaseEvent(event)

    def undo(self):
        if len(self.history) > 1:
            self.redo_list.append(self.history.pop().copy())
            # remove top item from history.
            new_state = self.history[-1].copy()
            self.annot_pixmap_holder.setPixmap(new_state)
            self.annot_pixmap = new_state
            self.parent.store_annot_slice()
            self.parent.parent.update_viewer_annot_slice()
            self.parent.parent.update_viewer_outline()

    def redo(self):
        if self.redo_list:
            new_state = self.redo_list.pop()
            self.history.append(new_state.copy())
            self.annot_pixmap_holder.setPixmap(new_state)
            self.annot_pixmap = new_state
            self.parent.store_annot_slice()
            # Update all views with new state.
            self.parent.parent.update_viewer_annot_slice()

    
    def update_axial_slice_pos_indicator(self):
        """ update the position of the axial slice indicator in the sagittal view """
        for v in self.parent.parent.viewers:
            if v.mode == 'sagittal':
                slice_nav = self.parent.parent.axial_viewer.slice_nav
                x1 = 0.0
                x2 = self.parent.parent.annot_data.shape[2]
                y1 = slice_nav.max_slice_idx - slice_nav.slice_idx
                y1 += 0.5
                y2 = y1
                if hasattr(v.scene, 'line'):
                   v.scene.line.setLine(x1, y1, x2, y2)
                else:
                    # add line if it doesn't exist
                    v.scene.line = QtWidgets.QGraphicsLineItem(x1, y1, x2, y2)
                    v.scene.line.setPen(QtGui.QPen(QtGui.QColor(255, 60, 60), 0.2,
                                        QtCore.Qt.DashLine))
                    v.scene.addItem(v.scene.line)
                    v.scene.line.setVisible(True)


    def flood_fill(self, x, y):
        raise Exception('disabled')
        x = round(x)
        y = round(y)
        # all rgb channels must match for the flood region to expand
        image =  self.annot_pixmap.toImage()
        rgb_np = np.array(qimage2ndarray.rgb_view(image))
        alpha_np = np.array(qimage2ndarray.alpha_view(image))
        mask = np.ones((rgb_np.shape[0], rgb_np.shape[1]), dtype=np.int)
        for i in range(3):
            mask *= flood(rgb_np[:, :, i], (y, x), connectivity=1)
        mask *= flood(alpha_np, (y, x), connectivity=1)
        
        # Assign previous labels to the new annotation before adding flooded region.        
        np_rgba = np.zeros((rgb_np.shape[0], rgb_np.shape[1], 4))
        fg = rgb_np[:, :, 0] > 0
        bg = rgb_np[:, :, 1] > 0
        np_rgba[:, :, 1] = bg * 255 # green is bg
        np_rgba[:, :, 0] = fg * 255 # red is fg
        np_rgba[:, :, 3] = (bg + fg) * 180 # alpha

        # flood fill should be with current brush color
        np_rgba[:, :, :][mask > 0] = [
            self.parent.brush_color.red(),
            self.parent.brush_color.green(),
            self.parent.brush_color.blue(),
            self.parent.brush_color.alpha()
        ]
        q_image = qimage2ndarray.array2qimage(np_rgba)
        return QtGui.QPixmap.fromImage(q_image)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier:
            # if control key is pressed then don't do anything 
            return
        self.mouse_down = True
        if (not modifiers & QtCore.Qt.ControlModifier
              and (self.parent.annot_visible or self.parent.outline_visible)
              and self.parent.parent.view_state == ViewState.ANNOTATING):
            pos = event.scenePos()
            x, y = pos.x(), pos.y()
            self.parent.parent.log(f'mouse_press,drawing,x:{x},y:{y}'
                                   f',brush_size:{self.brush_size}'
                                   f',brush_color:{self.parent.brush_color.name()}'
                                   f',mode:{self.parent.mode}')

            if modifiers == (QtCore.Qt.AltModifier | QtCore.Qt.ShiftModifier):
                idx = self.parent.slice_nav.max_slice_idx - self.parent.cur_slice_idx 
                self.regions_to_restrict_to.append((round(x), round(y), idx))
                self.mouse_down = False
            #elif modifiers == QtCore.Qt.AltModifier:
                # if alt key is pressed then we want to flood fill
                #  from the clicked region.
            #     self.annot_pixmap = self.flood_fill(x, y)
            #     self.annot_pixmap_holder.setPixmap(self.annot_pixmap)
            else:
                # draw the circle
                if self.brush_size == 1:
                    circle_x = x
                    circle_y = y
                else:
                    circle_x = x - (self.brush_size / 2) + 0.5
                    circle_y = y - (self.brush_size / 2) + 0.5

                painter = QtGui.QPainter(self.annot_pixmap)
                painter.setCompositionMode(QtGui.QPainter.CompositionMode_Source)
                painter.drawPixmap(0, 0, self.annot_pixmap)
                painter.setPen(QtGui.QPen(self.parent.brush_color, 0, Qt.SolidLine,
                                          Qt.RoundCap, Qt.RoundJoin))
                painter.setBrush(QtGui.QBrush(self.parent.brush_color, Qt.SolidPattern))            
                if self.brush_size == 1:
                    painter.drawPoint(round(circle_x), round(circle_y))
                else:
                    painter.drawEllipse(round(circle_x), round(circle_y), round(self.brush_size-1), round(self.brush_size-1))
                self.annot_pixmap_holder.setPixmap(self.annot_pixmap)
                painter.end()
            self.last_x = x
            self.last_y = y


    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)

        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier:
            # if control key is pressed then don't do anything with the graphics scene
            return

        if self.parent.parent.view_state == ViewState.ANNOTATING:
            self.parent.parent.log(f'mouse_release,drawing,mode:{self.parent.mode}')
            # has to be some limit to history or RAM will run out
            if len(self.history) > 50:
                self.history = self.history[-50:]
            self.history.append(self.annot_pixmap.copy())
            self.redo_list = []
            self.parent.store_annot_slice()
            # update all views with new state.
            self.parent.parent.update_viewer_annot_slice()
            self.parent.parent.update_viewer_outline()

            pos = event.scenePos()
            x, y = pos.x(), pos.y()
            idx = self.parent.slice_nav.max_slice_idx - self.parent.cur_slice_idx            
            
            if self.parent.parent.auto_complete_enabled:
                prev_annot = np.array(qimage2ndarray.rgb_view(self.history[-2].toImage()))
                recent_annot = np.array(qimage2ndarray.rgb_view(self.history[-1].toImage()))
                diff = np.absolute(np.array(prev_annot) - np.array(recent_annot))
                if np.any(diff):
                    yy, xx, __ = np.where(diff > 0)
                    centroid_y = int(round(np.mean(yy)))
                    centroid_x = int(round(np.mean(xx)))
                    self.patch_segmentor.segment_patch(round(centroid_x), round(centroid_y), idx)
         
        self.mouse_down = False


    def clear_cursor(self):
        if self.cursor_shown:
            self.cursor_pixmap.fill(Qt.transparent) 
            self.cursor_pixmap_holder.setPixmap(self.cursor_pixmap)
            self.cursor_shown = False

    def update_cursor(self, ctrl_press=None):
        # implemented for changing brush color of static cursor
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier or ctrl_press:
            self.clear_cursor()
        elif self.cursor_shown and hasattr(self, 'cursor_x'):
            self.drawPaintCursorXY(self.cursor_x, self.cursor_y)

    def drawPaintCursor(self, event):
        # code from the mousedown draw function
        pos = event.scenePos()
        x, y = pos.x(), pos.y()
        self.drawPaintCursorXY(x, y)

    def drawPaintCursorXY(self, x, y):
        self.cursor_x = x
        self.cursor_y = y
        self.clear_cursor()
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if (not modifiers == QtCore.Qt.ControlModifier and
            self.parent.parent.view_state == ViewState.ANNOTATING):
            if self.brush_size == 1:
                circle_x = x
                circle_y = y
            else:
                circle_x = x - (self.brush_size / 2) + 0.5
                circle_y = y - (self.brush_size / 2) + 0.5
            painter = QtGui.QPainter(self.cursor_pixmap)
            painter.setCompositionMode(QtGui.QPainter.CompositionMode_Source)
            painter.drawPixmap(0, 0, self.cursor_pixmap)
            color = QtGui.QColor.fromRgbF(self.parent.brush_color.red(),
                                          self.parent.brush_color.green(),
                                          self.parent.brush_color.blue(), 0.4)
            painter.setPen(QtGui.QPen(color, 0, Qt.SolidLine,
                                      Qt.RoundCap, Qt.RoundJoin))
            painter.setBrush(QtGui.QBrush(color, Qt.SolidPattern))

            if self.brush_size == 1:
                painter.drawPoint(round(circle_x), round(circle_y))
            else:
                painter.drawEllipse(round(circle_x), round(circle_y), round(self.brush_size-1), round(self.brush_size-1))
            painter.end()
            self.cursor_shown = True
            self.cursor_pixmap_holder.setPixmap(self.cursor_pixmap)

    def update_info_label(self, x, y):
        try:
            idx = (self.parent.slice_nav.max_slice_idx - self.parent.cur_slice_idx) - 1
            x = round(x)
            y = round(y)
            x = max(0, x)
            y = max(0, y)
            x = min(self.parent.parent.img_data.shape[2]-1, x)
            y = min(self.parent.parent.img_data.shape[1]-1, y)
            hu = self.parent.parent.img_data[idx, y, x]
            self.parent.parent.hu_label.setText(f'HU: {hu} ') # | Position: (x:{round(x)}, y:{round(y)}')
        except Exception as _:
            pass

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        self.drawPaintCursor(event)
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        shift_down = (modifiers & QtCore.Qt.ShiftModifier)

        pos = event.scenePos()
        x, y = pos.x(), pos.y()

        if modifiers == QtCore.Qt.ControlModifier:
            # if control key is pressed then don't do anything with the graphics scene
            return

        elif shift_down:
            dist = self.last_y - y
            self.brush_size += dist
            self.brush_size = max(1, self.brush_size)
            # Warning: very tight coupling.
            self.parent.update_cursor()

        elif (self.parent.parent.view_state == ViewState.ANNOTATING
              and self.mouse_down and (self.parent.annot_visible or self.parent.outline_visible)):

            painter = QtGui.QPainter(self.annot_pixmap)
            painter.setCompositionMode(QtGui.QPainter.CompositionMode_Source)
            painter.drawPixmap(0, 0, self.annot_pixmap)
            pen = QtGui.QPen(self.parent.brush_color, self.brush_size, Qt.SolidLine,
                             Qt.RoundCap, Qt.RoundJoin)
            painter.setPen(pen)
            pos = event.scenePos()
            x, y = pos.x(), pos.y()

            # Based on empirical observation
            if self.brush_size % 2 == 0:
                painter.drawLine(self.last_x+0.5, self.last_y+0.5, x+0.5, y+0.5)
            else:
                painter.drawLine(round(self.last_x), round(self.last_y), round(x), round(y))

            self.annot_pixmap_holder.setPixmap(self.annot_pixmap)
            painter.end()
        self.update_info_label(x, y) 

        self.last_x = x
        self.last_y = y



"""
Container for canvas where image and lines are drawn.
Facilitates use of zoom and pan.

Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

#pylint: disable=I1101,E0611,C0111
from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5.QtCore import Qt

class CustomGraphicsView(QtWidgets.QGraphicsView):
    """
    Container for canvas where image and lines are drawn.
    Facilitates use of zoom and pan.
    """
    mouse_scroll_event = QtCore.pyqtSignal(QtGui.QWheelEvent)
    zoom_change = QtCore.pyqtSignal()
    leave = QtCore.pyqtSignal()
    enter = QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        self.zoom = 1
        self.setDragMode(QtWidgets.QGraphicsView.NoDrag)

    def update_zoom(self):
        """ Transform the view based on current zoom value """
        self.setTransform(QtGui.QTransform().scale(self.zoom, self.zoom))
        self.zoom_change.emit()

    def wheelEvent(self, event):
        self.mouse_scroll_event.emit(event)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
        super().keyPressEvent(event)

    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
        super().keyReleaseEvent(event)

    def show_actual_size(self):
        self.zoom = 1
        self.update_zoom()

    def leaveEvent(self, event):
        self.leave.emit()

    def enterEvent(self, event):
        self.enter.emit()

    def fit_to_view(self, _event=None):
        view_width = self.geometry().width()
        view_height = self.geometry().height()
        im_width = self.image.width()
        im_height = self.image.height()
        width_ratio = im_width / view_width
        height_ratio = im_height / view_height

        if width_ratio > height_ratio:
            self.zoom = view_width / im_width
        else:
            self.zoom = view_height / im_height

        scene_rect = self.sceneRect()
        def fitin():
            self.fitInView(scene_rect, Qt.KeepAspectRatio)
        fitin()
        QtCore.QTimer.singleShot(100, fitin)
        self.zoom_change.emit()



""" 
Copyright (C) 2022 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from paramiko import SSHClient
import paramiko 
from scp import SCPClient
import os
import numpy as np

scp_client = None

def scp_transfer(seg, seg_fname, remote_ip, remote_uname):
    global scp_client 
    if not scp_client:
        ssh = SSHClient()
        ssh.load_system_host_keys()
        # be careful with this - private network only
        # ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(remote_ip, username=remote_uname)
        # SCP Client takes a paramiko transport as an argument
        scp_client = SCPClient(ssh.get_transport())
    tmp_out = os.path.join('tmp', seg_fname)
    np.savez_compressed(tmp_out, seg=seg)
    scp_client.put(tmp_out, 'rp_scp_in')
